﻿
' Module which controls the driving of the robot. 

Imports System.Math
Imports System.Threading


Module modBotMovementControl

    Dim tmcLastLocalizedPosition As typeMapCoordinate
    Dim intConsecutiveTurnCounter As Short = 0  ' Counts the consecutive number of "Turn" moves.
    '                                                  Used to prevent excessive localization and new path generation 


    Public Sub subInitializeDrivingControl(ByVal tbpEstimatedPose As typeBotPose)
        If Not boolDrivingToDestination Then 'initialize
            intCollisions = 0
            boolBotIsMoving = False
            GotABotClass.subBotSpeedCalculations()
            boolPoseConfirmedAfterCollision = True ' 
            tmcLastLocalizedPosition = tbpEstimatedPose.Location
            intMovementsSinceLocalization = 0
            ftnLocalizeUsingOdometry(tbpEstimatedPose, tbpStartingPose) ' initialize tmcStartingOdometry
            intConsecutiveTurnCounter = 0
        End If
    End Sub


    Public Function ftnBotStop(ByVal tbpEstimatedPose As typeBotPose) As typeBotPose
        On Error Resume Next
        strStatusDriving = intDrivingSequenceNumber & " : Stop: "
        ftnBotStop = ftnExecuteMove(tbpEstimatedPose, "stop", tbpEstimatedPose)
        GotABotClass.subUpdateStatusLine()
    End Function


    Function ftnBotFwd(ByVal tbpEstimatedPose As typeBotPose, ByVal tbpTargetPose As typeBotPose) As typeBotPose
        On Error Resume Next
        strTypeOfMove = "Fwd"
        Dim sngFwdDistance As Single = ftnDistanceCommanded(tbpEstimatedPose, tbpTargetPose)
        If sngFwdDistance = 0 Then
            sngFwdDistance = sngTwoInches 'Error Recovery
            tbpTargetPose.Location = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngFwdDistance)
            frmDebug.subWriteLogs(Color.Red, Now() & " ***** Path Failure:  Fwd 2 commanded ***************************")
        ElseIf (sngFwdDistance > sngMaxTravel) Then
            sngFwdDistance = sngMaxTravel
            tbpTargetPose.Location = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngFwdDistance)
        End If
#If constCompile2dAccelerometer Then
ftnTransmitGotAKinect("get accelerometer distance ", "OK", 1500) ' start GotAKinects distance accumulation
#End If
        intConsecutiveTurnCounter = 0
        If strUnits = "inch" Then
            ftnBotFwd = ftnExecuteMove(tbpEstimatedPose, "move " & sngFwdDistance & " i", tbpTargetPose)
        Else
            ftnBotFwd = ftnExecuteMove(tbpEstimatedPose, "move " & sngFwdDistance & " cm", tbpTargetPose)
        End If
    End Function

    Function ftnBotXY(ByVal tbpEstimatedPose As typeBotPose, ByVal tbpTargetPose As typeBotPose) As typeBotPose
        On Error Resume Next
        strTypeOfMove = "XY"
        Dim tmcXY_Distances As typeMapCoordinate = ftnFindXY_Parameters(tbpEstimatedPose, tbpTargetPose)
        Dim sngFwdDistance As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tbpTargetPose.Location)
        If sngFwdDistance = 0 Then
            frmDebug.subWriteLogs(Color.Red, Now() & " ***** Path Failure - move XY:  move  2  i commanded ***************************")
            tbpTargetPose.Location = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngTwoInches)
            ftnBotXY = ftnBotFwd(tbpEstimatedPose, tbpTargetPose)
            Exit Function
        ElseIf (sngFwdDistance > sngMaxTravel) Then
            ftnBotXY = ftnBotTurn(tbpEstimatedPose, True, tbpTargetPose)
            Exit Function
        End If

        'tbpTargetPose.Orientation = ftnConvertTo360(ftnTurnAngle(tbpEstimatedPose, tbpTargetPose.Location))
        intConsecutiveTurnCounter = 0
        If strUnits = "inch" Then
            ftnBotXY = ftnExecuteMove(tbpEstimatedPose, "move X " & Int(tmcXY_Distances.X) & " Y " & Int(tmcXY_Distances.Y) & " i", tbpTargetPose)
        Else
            ftnBotXY = ftnExecuteMove(tbpEstimatedPose, "move X " & Int(tmcXY_Distances.X) & " Y " & Int(tmcXY_Distances.Y) & " cm", tbpTargetPose)
        End If
    End Function

    Function ftnBotTurn(ByVal tbpEstimatedPose As typeBotPose, ByVal boolRestrictConsecutiveTurns As Boolean, _
                         ByVal tbpTargetPose As typeBotPose) As typeBotPose
        On Error Resume Next
        strTypeOfMove = "Turn"
        If boolRestrictConsecutiveTurns And (intConsecutiveTurnCounter >= constMaxTurnsWithLocalization) Then ' intConsecutiveTurnCounter keeps track of turns and small movements and is used to limit unproductive localizations and path regeneration
            strTypeOfMove = "Fwd"
            frmDebug.subWriteLogs(Color.Red, Now() & " *********** Too many consecutive Turns - Moving forward instead ***************************")
            tbpTargetPose = tbpEstimatedPose
            tbpTargetPose.Location = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngTwoInches)
            ftnBotTurn = ftnBotFwd(tbpEstimatedPose, tbpTargetPose)
        Else
            Dim sngTurnAngle As Single = ftnAngleDifference(tbpTargetPose.Orientation, tbpEstimatedPose.Orientation)
            If (Abs(sngTurnAngle) > 135) Then sngTurnAngle = ftnTurnWithoutHittingTail(tbpEstimatedPose, sngTurnAngle)
            ftnBotTurn = ftnExecuteMove(tbpEstimatedPose, "move " & sngTurnAngle & " degrees", tbpTargetPose)
            intConsecutiveTurnCounter += 1
        End If
    End Function

    Function ftnTurnWithoutHittingTail(ByVal tbpEstimatedPose As typeBotPose, ByVal sngTurnAngle As Single) As Single
        ' This routine assumes that the robot is not perfectly round but rather that the robot's tail extends
        ' furthur from the center of movement than the head or sides.  To lessen the chances of collision between
        ' the tail and a nearby wall while turning, this routine will reverse the turn direction when near a wall 
        ' and when the direction of turn moves the tail towards the wall.
        On Error Resume Next
        ftnTurnWithoutHittingTail = sngTurnAngle
        Dim boolNearWallLeft As Boolean = False
        Dim boolNearWallRight As Boolean = False
        Dim boolNearWallRear As Boolean = False
        Dim tmcCheckLocation As typeMapCoordinate
        tmcCheckLocation = ftnLocationFromLocation(tmcCheckLocation, ftnConvertTo360(tbpEstimatedPose.Orientation - 180), sngBotRadius)
        boolNearWallRear = boolOccupiedCell(tmcCheckLocation.X, tmcCheckLocation.Y)
        If Not boolNearWallRear Then
            Dim sngDistance As Single = sngBotRadius
            While Not boolNearWallLeft And Not boolNearWallRight And (sngDistance < (sngBotRadius + sngSixInches)) ' Find if there is a wall near the sides of the bot
                tmcCheckLocation = ftnLocationFromEstimatedPose(tbpEstimatedPose, 90, sngDistance)
                boolNearWallLeft = boolOccupiedCell(tmcCheckLocation.X, tmcCheckLocation.Y)
                tmcCheckLocation = ftnLocationFromEstimatedPose(tbpEstimatedPose, -90, sngDistance)
                boolNearWallRight = boolOccupiedCell(tmcCheckLocation.X, tmcCheckLocation.Y)
                sngDistance += sngTwoInches
            End While
            If boolNearWallLeft And Not boolNearWallRight And (sngTurnAngle < 0) Then
                ftnTurnWithoutHittingTail = 360 + sngTurnAngle
            ElseIf boolNearWallRight And Not boolNearWallLeft And (sngTurnAngle > 0) Then
                ftnTurnWithoutHittingTail = -(360 - sngTurnAngle)
            Else
                ftnTurnWithoutHittingTail = sngTurnAngle
            End If
        End If
    End Function

    Dim boolCourseCorrected As Boolean = False
    Function ftnExecuteMove(ByVal tbpEstimatedPose As typeBotPose, ByVal strCommand As String, ByVal tbpTargetPose As typeBotPose) As typeBotPose ' Execute move and update pose using dead reckoning 
        On Error Resume Next
        ftnExecuteMove = tbpEstimatedPose
        boolRangeReadingsAreFresh = False
        If boolSimulatedMovement Then
            ftnExecuteMove = ftnSimulatedExecuteMove(tbpEstimatedPose, strCommand, tbpTargetPose)
        ElseIf Not boolShutDown Then
            If (boolStop Or (strCommand = "stop")) Then   'And boolBotIsMoving
                boolCourseCorrected = True
                ftnExecuteMove = ftnLocalizeUsingDeadReckoning(tbpEstimatedPose, 0, tbpTargetPose)
                Dim strResponse As String = ftnSendCommand("stop")
                boolBotIsMoving = False
                frmDebug.WritePose(ftnElapsedTime() & "       Pose Derived From Dead Reckoning", ftnExecuteMove)
                tbpTargetPose = ftnExecuteMove
                ftnExecuteMove = ftnLocalizeUsingOdometry(ftnExecuteMove, tbpStartingPose)
                ftnExecuteMove = ftnCheckForMoveComError(ftnExecuteMove, strResponse, tbpTargetPose)
                tbpTargetPoseGlobal = ftnExecuteMove
                tbpStartingPose = ftnExecuteMove
                ftnExecuteMove.LocationUncertainty.X += sngFiveInches / sngGridSize
                ftnExecuteMove.LocationUncertainty.Y += sngFiveInches / sngGridSize
                ftnExecuteMove.OrientationUncertainty += 8
                intMovementsSinceLocalization = 99  ' force localization 
                frmDebug.subWriteLogs(Color.Red, ftnElapsedTime() & " ************************ STOP commanded  ********************************")
                If boolCollisionPredicted Then
                    subDrawScanVectors(tbpBotPose, False)
                    frmDebug.subWriteLogs(Color.Red, "                (Collision was predicted)")
                    ftnSendCommand("play phrase 'Emergency Stop.  '")
                    boolCollisionPredicted = False
                End If
                boolCourseCorrected = False
            ElseIf (strTypeOfMove = "Fwd") Or (strTypeOfMove = "XY") Or (strTypeOfMove = "Turn") And Not boolStop Then
                tbpTargetPoseGlobal = tbpTargetPose
                Dim boolMidCourseCorrectionCommand As Boolean = boolBotIsMoving
                If Not boolMidCourseCorrectionCommand Then  ' New move / not a course correction
                    If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, "          <<<<<  ftnExecuteMove  START     ftnElapsedTime() = " & ftnElapsedTime())
                    ftnCheckBatteryForDisconnect() '   
                    boolCourseCorrected = False
                    frmDebug.subWriteMoveStartLogFile(strCommand, tbpTargetPose)
                    strStatusDriving = intDrivingSequenceNumber & " : Driving: " & strCommand
                    GotABotClass.subUpdateStatusLine()
                    tbpStartingPose = tbpEstimatedPose
                    If (Abs(ftnAngleDifference(tbpTargetPose.Orientation, tbpEstimatedPose.Orientation)) >= 25) Then intMovementsSinceLocalization = 99 ' Force localization
                    timeMoveCommandSent = Now()
                    Dim strResponse As String = ftnSendCommand(strCommand)
                    boolBotIsMoving = False
                    ftnExecuteMove = ftnLocalizeUsingDeadReckoning(ftnExecuteMove, 0, tbpTargetPoseGlobal)
                    frmDebug.WritePose(ftnElapsedTime() & "       Pose Derived From Dead Reckoning", ftnExecuteMove)
                    If boolCourseCorrected Then
                        ftnLocalizeUsingOdometry(ftnExecuteMove, tbpStartingPose)
                    Else
                        ftnExecuteMove = ftnLocalizeUsingOdometry(ftnExecuteMove, tbpStartingPose)
                    End If
                    ftnExecuteMove = ftnCheckForMoveComError(ftnExecuteMove, strResponse, tbpTargetPose)
                    tbpTargetPose = ftnExecuteMove
                    tbpStartingPose = ftnExecuteMove
                    intDrivingSequenceNumber += 1
                    If boolEnableLocalizationWhileMoving Then subEndImageRecognition() '
                    boolRangeReadingsAreFresh = False
                    If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, "          <<<<<  ftnExecuteMove  END     ftnElapsedTime() = " & ftnElapsedTime())
                Else
                    boolCourseCorrected = True
                    If boolDebugLWM Then strDebug &= ftnElapsedTime() & " $$$$$$$$$$$$$$$$$$$$$$$$ Course Correction command:  " & strCommand & vbCrLf
                    Dim strResponse As String = ftnSendCommand(strCommand)
                End If
            End If
        End If
    End Function


    Public Function ftnPostMovement(ByVal tbpEstimatedPose As typeBotPose, ByVal strTypeOfMove As String, _
                                    ByVal tbpStartingPose As typeBotPose, ByVal tbpTargetPose As typeBotPose) As typeBotPose
        'localizes, finds new path if required, collision recovery if required
        On Error Resume Next
        If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, "          <<<<<  ftnPostMovement   START    ftnElapsedTime() = " & ftnElapsedTime())
        ftnPostMovement = tbpEstimatedPose
        If boolSimulatedMovement Then
            If (boolCollisionDetected) Then
                ftnPostMovement = ftnCollisionRecovery(ftnPostMovement)
            Else
                ftnPostMovement.LocationUncertainty = tbpStartingPose.LocationUncertainty
                ftnPostMovement.OrientationUncertainty = tbpStartingPose.OrientationUncertainty
                If (intPathArray(ftnPostMovement.Location.X, ftnPostMovement.Location.Y) < constDestinationCost) Then  ' only required if the Bot is too far off the path
                    If boolSlam Then
                        ftnFindSlamPath()
                    Else
                        subGeneratePathMatrix()
                        ftnGeneratePathCosts(tmcDestination, ftnPostMovement.Location)
                    End If
                End If
            End If
        ElseIf Not boolShutDown Then
            If (tbpEstimatedPose.OrientationUncertainty < 60) Then ftnPostMovement = frmCompass.ftnLocalizeUsingCompass(ftnPostMovement, tbpTargetPose) ' Don't use Compass during Lost Localization
            If (boolCollisionDetected) Then
                ftnPostMovement = ftnCollisionRecovery(ftnPostMovement)
                boolRangeReadingsAreFresh = False
            Else
                boolRangeReadingsAreFresh = False
                If ftnLocalizationNeeded(tbpEstimatedPose, tbpStartingPose) Then ' Fewer localizations result in fewer driving corrections, but also greater chance of collisions - its a tradeoff
                    ftnPostMovement = ftnLocalize_Top(ftnPostMovement)
                    If (ftnPoseUncertaintyFactor(ftnPostMovement) < ftnPoseUncertaintyFactor(tbpEstimatedPose)) Then tmcLastLocalizedPosition = ftnPostMovement.Location
                End If
            End If
            If ftnAmIStuck(ftnPostMovement, strTypeOfMove, ftnAngleDifference(tbpStartingPose.Orientation, tbpTargetPose.Orientation), tbpStartingPose) Then _
                ftnPostMovement = ftnGetUnStuck(ftnPostMovement, strTypeOfMove)
            ftnPostMovement = ftnLostRecoveryTop(ftnPostMovement, tbpTargetPose)
            If (intPathArray(ftnPostMovement.Location.X, ftnPostMovement.Location.Y) < constDestinationCost) Then
                subGeneratePathMatrix() ' Bot is too far off the original path, must regenerate path
                ftnGeneratePathCosts(tmcDestination, ftnPostMovement.Location)
            End If
            tbpTargetPoseGlobal = ftnPostMovement
        End If
        If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, "          <<<<<  ftnPostMovement   END    ftnElapsedTime() = " & ftnElapsedTime())
    End Function

    Function ftnCheckForMoveComError(ByVal tbpEstimatedPose As typeBotPose, ByVal strResponse As String, ByVal tbpTargetPose As typeBotPose) As typeBotPose
        On Error Resume Next
        ftnCheckForMoveComError = tbpEstimatedPose
        If CBool(InStr(strResponse, "com error")) Then ' some type of error was detected so increase uncertainties
            If (strTypeOfMove = "Fwd") Or (strTypeOfMove = "XY") Then
                Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tbpTargetPose.Location)
                tbpStartingPose.LocationUncertainty.X += sngDistanceToTarget / 2 / sngGridSize
                tbpStartingPose.LocationUncertainty.Y += sngDistanceToTarget / 2 / sngGridSize
            End If
            If (strTypeOfMove = "Turn") Or (strTypeOfMove = "XY") Then tbpStartingPose.OrientationUncertainty += Abs(ftnAngleDifference(tbpEstimatedPose.Orientation, tbpTargetPose.Orientation) / 2)
            ftnCheckForMoveComError = tbpStartingPose
            ftnSendCommand("play phrase 'Communication error.  '")
            ftnCheckForMoveComError.LocationUncertainty.X = ftnMaxValue(ftnCheckForMoveComError.LocationUncertainty.X, 12)
            ftnCheckForMoveComError.LocationUncertainty.Y = ftnMaxValue(ftnCheckForMoveComError.LocationUncertainty.Y, 12)
            ftnCheckForMoveComError.OrientationUncertainty = ftnMaxValue(ftnCheckForMoveComError.OrientationUncertainty, 17)
            intMovementsSinceLocalization = 4  ' force compass and localization updates
            frmDebug.subWriteLogs(Color.Red, "  ")
            frmDebug.subWriteLogs(Color.Red, Now() & " **************  Communication Error during movement.  ' *****************************")
            frmDebug.subWriteLogs(Color.Red, "  ")
        ElseIf CBool(InStr(strResponse, "com timeout")) Then ' some type of error was detected so increase uncertainties
            ftnCheckForMoveComError.LocationUncertainty.X = ftnMaxValue(ftnCheckForMoveComError.LocationUncertainty.X, 12)
            ftnCheckForMoveComError.LocationUncertainty.Y = ftnMaxValue(ftnCheckForMoveComError.LocationUncertainty.Y, 12)
            ftnCheckForMoveComError.OrientationUncertainty = ftnMaxValue(ftnCheckForMoveComError.OrientationUncertainty, 17)
            intMovementsSinceLocalization = 4  ' force compass and localization updates
        End If
    End Function

    Function ftnSimulatedExecuteMove(ByVal tbpEstimatedPose As typeBotPose, ByVal strCommand As String, ByVal tbpTargetPose As typeBotPose) As typeBotPose
        On Error Resume Next
        ftnSimulatedExecuteMove = tbpEstimatedPose
        tbpTargetPoseGlobal = tbpTargetPose
        frmDebug.subDebugLog(Color.Black, ">: " & strCommand)
        Dim sngStartOfMoveDelayOld As Single = sngStartOfMoveDelay
        sngStartOfMoveDelay = 0
        subDoEventsIfMainThread()
        timeMoveCommandSent = Now()
        If (strCommand = "stop") Then
            tbpTargetPoseGlobal = tbpEstimatedPose
            strTypeOfMove = ""
        Else
            tbpStartingPose = tbpEstimatedPose
            Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tbpTargetPose.Location)
            frmDebug.subWriteMoveStartLogFile(strCommand, tbpTargetPose)
            While ((strTypeOfMove = "Turn") And ((Abs(ftnConvertTo180(ftnAngleDifference(tbpEstimatedPose.Orientation, tbpTargetPose.Orientation))) - 0.5) >= Abs(ftnAngleDifference(tbpStartingPose.Orientation, ftnSimulatedExecuteMove.Orientation))) Or _
                 (strTypeOfMove = "Fwd") And ((Abs(sngDistanceToTarget) - sngGridSize) >= ftnDistanceBetweenLocations(tbpStartingPose.Location, ftnSimulatedExecuteMove.Location)) Or _
                 (strTypeOfMove = "XY") And ((Abs(sngDistanceToTarget) - sngGridSize) >= ftnDistanceBetweenLocations(tbpStartingPose.Location, ftnSimulatedExecuteMove.Location))) And _
                 Not boolStop And Not boolShutDown
                boolBotIsMoving = True
                Dim tbpOldLocation As typeMapCoordinate = ftnSimulatedExecuteMove.Location
                ftnSimulatedExecuteMove = ftnLocalizeUsingDeadReckoning(ftnSimulatedExecuteMove, 0, tbpTargetPose)
                subEraseOldBotLocation(tbpOldLocation)
                subDrawBot(ftnSimulatedExecuteMove)
                subDoEventsIfMainThread()
            End While
        End If
        tbpStartingPose = ftnSimulatedExecuteMove
        intDrivingSequenceNumber += 1
        tbpTargetPoseGlobal = ftnSimulatedExecuteMove
        boolBotIsMoving = False
        boolCollisionPredicted = False
        sngStartOfMoveDelay = sngStartOfMoveDelayOld
        subDoEventsIfMainThread()
    End Function



    Public Function ftnLocalizationNeeded(ByVal tbpEstimatedPose As typeBotPose, ByVal tbpStartingPose As typeBotPose) As Boolean
        On Error Resume Next
        ftnLocalizationNeeded = False
        If boolSimulatedMovement Or boolShutDown Then
            ftnLocalizationNeeded = False
        ElseIf (ftnPoseUncertaintyFactor(tbpStartingPose) > 480) Then '(15+15)*16
            ftnLocalizationNeeded = True
        ElseIf boolSlam Then
            ftnLocalizationNeeded = True
        ElseIf Not boolPoseConfirmedAfterCollision Then
            ftnLocalizationNeeded = True
        ElseIf (ftnDistanceBetweenLocations(tmcLastLocalizedPosition, tbpEstimatedPose.Location) > sngSixInches) Then
            ftnLocalizationNeeded = True
        ElseIf (intMovementsSinceLocalization > 9) Then
            ftnLocalizationNeeded = True
        ElseIf (intConsecutiveTurnCounter <= constMaxTurnsWithLocalization) Then ' This prevents Bot from turning back and forth repeatedly
            If (intMovementsSinceLocalization > 2) Then
                ftnLocalizationNeeded = True
                'ElseIf (Abs(ftnAngleDifference(tbpStartingPose.Orientation, tbpEstimatedPose.Orientation)) >= 25) Then  ' doesn't work when LWM updates the starting pose
                '    ftnLocalizationNeeded = True
            End If
        End If
        If ftnLocalizationNeeded Then
            intMovementsSinceLocalization = 0
        Else
            intMovementsSinceLocalization += 1
        End If
    End Function


    Function ftnCheckBatteryForDisconnect() As Boolean
        On Error Resume Next
        ftnCheckBatteryForDisconnect = False
        If Not boolSimulatedMovement Then
            Do
                ftnSendCommand("battery") '   
                If (strStatusBotVoltage = "   ??   ") Or CBool(InStr(strStatusBotVoltage, "--")) Or (sngBotBatteryVoltage < (sngDockTriggerBot * 0.7)) Then
                    If Not boolMute Then Beep()
                    ftnCheckBatteryForDisconnect = True
                    GotABotClass.StatusDriving.ForeColor = Color.Red
                    strStatusDriving = intDrivingSequenceNumber & " : BATTERY DISCONNECTED! "
                    GotABotClass.subUpdateStatusLine()
                    subWait(1000)
                    frmDebug.subWriteLogs(Color.Black, Now() & "    |||||||||||  BATTERY DISCONNECT DETECTED  |||||||||||")
                    frmDebug.subWriteLogs(Color.Black, Now() & "                   Battery Voltage = " & sngBotBatteryVoltage)
                    subDoEventsIfMainThread()
                    If Not boolMute Then Beep()
                    subWait(1500)
                    ftnSendCommand("battery") '   
                End If
            Loop Until (strStatusBotVoltage <> "   ??   ") Or Not CBool(InStr(strStatusBotVoltage, "--")) Or (sngBotBatteryVoltage > (sngDockTriggerBot * 0.7)) Or boolStop Or boolShutDown
            If ftnCheckBatteryForDisconnect Then
                frmDebug.subWriteLogs(Color.Black, Now() & "                   Battery Reconnected.  Voltage = " & sngBotBatteryVoltage)
            End If
        End If
    End Function

    Function ftnAmIStuck(ByVal tbpEstimatedPose As typeBotPose, ByVal strTypeOfMove As String, ByVal sngMoveDistance As Single, ByVal tbpStartingPose As typeBotPose) As Boolean
        On Error Resume Next
        ftnAmIStuck = False
        If (strTypeOfMove = "Turn") Then
            If (Abs(ftnAngleDifference(tbpStartingPose.Orientation, tbpEstimatedPose.Orientation)) < (Abs(ftnConvertTo180(sngMoveDistance)) - tbpStartingPose.OrientationUncertainty - tbpEstimatedPose.OrientationUncertainty)) Then
                ftnAmIStuck = True
                ftnSendCommand("play phrase 'I am stuck'")
                frmDebug.subWriteLogs(Color.Black, Now() & "    |||||||||||  I'm stuck  |||||||||||")
            End If
        End If
    End Function

    Public Function ftnGetUnStuck(ByVal tbpEstimatedPose As typeBotPose, ByVal strTypeOfMove As String) As typeBotPose
        On Error Resume Next
        ftnGetUnStuck = tbpEstimatedPose
        If Not boolStop And Not boolShutDown Then
            subDoEventsIfMainThread()
            If (strTypeOfMove = "Turn") Then
                tbpTargetPoseGlobal.Location = ftnLocationFromEstimatedPose(tbpBotPose, 0, sngSixInches)
                tbpBotPose = ftnBotFwd(tbpBotPose, tbpTargetPoseGlobal)
                tbpTargetPoseGlobal.Location = ftnLocationFromEstimatedPose(tbpBotPose, 180, sngTwoInches)
                tbpBotPose = ftnBotFwd(tbpBotPose, tbpTargetPoseGlobal)
            ElseIf (strTypeOfMove = "Fwd") Then
                tbpTargetPoseGlobal.Location = ftnLocationFromEstimatedPose(tbpBotPose, 180, sngSixInches)
                tbpBotPose = ftnBotFwd(tbpBotPose, tbpTargetPoseGlobal)
                tbpTargetPoseGlobal.Location = ftnLocationFromEstimatedPose(tbpBotPose, 0, sngTwoInches)
                tbpBotPose = ftnBotFwd(tbpBotPose, tbpTargetPoseGlobal)
            End If
            subDoEventsIfMainThread()
        End If
        ftnGetUnStuck = tbpBotPose
    End Function

End Module
