﻿
Imports System.Math

Module modDriveToDestination

    ' This module determines the driving commands required for the Bot
    ' to reach a specific destination

    Public Sub subDriveToDestination_Top(ByVal tmcDestination As typeMapCoordinate)
        On Error Resume Next
        frmDebug.subWriteLogs(Color.Black, Now() & " Drive to new destination: " & tmcDestination.X & ", " & tmcDestination.Y)
        If Not boolSlam Then intDrivingSequenceNumber = 0
        If boolBotIsMoving Then tbpBotPose = ftnBotStop(tbpBotPose)
        subInitializeDrivingControl(tbpBotPose)  ' initialize
        intLastCollisionSequenceNumber = -6
        boolCollisionDetectionEnable = True
        If boolStop Then
            MsgBox(" Please exit STOP mode first")
            Exit Sub
        End If
        subCheckForPause()
        timeDrivingStarted = Now()
        If Not boolSlam Then ftnSendCommand("play phrase 'OK'")
        boolDrivingToDestination = True
        boolRangeReadingsAreFresh = False
        tbpStartingPose = tbpBotPose
        frmDock.subUnDock()
        subDrawDestination(tmcDestination)
        If boolSlam Then
            ftnFindSlamPath()
        Else
            subGeneratePathMatrix()
            ftnGeneratePathCosts(tmcDestination, tbpBotPose.Location)
        End If
        subDriveToDestinationLoop(tmcDestination)
        subSpeakDestinationSuccess(tmcDestination)
        ftnSendCommand("stop")
        If Not GotABotClass.mnuDock.Checked And Not GotABotClass.InvokeRequired And Not boolDock Then
            frmDebug.subWriteFinishedMessage()
            GotABotClass.StatusDriving.ForeColor = Color.Black
            strStatusDriving = intDrivingSequenceNumber & " : Finished Driving "
        End If
        GotABotClass.subRedraw(tbpBotPose)
        boolDrivingToDestination = False
    End Sub

    Sub subDriveToDestinationLoop(ByVal tmcDestination As typeMapCoordinate)
        On Error Resume Next
        Dim boolGoodPath As Boolean = True
        Dim tbpTargetPose As typeBotPose
        Do While ((ftnDistanceBetweenLocations(tbpBotPose.Location, tmcDestination) > sngBotDiameter * 0.33333) _
                         And Not boolStop And boolGoodPath _
                         And Not (boolSlam And ((byteSlamRawMapArray(tmcDestination.X, tmcDestination.Y) <> 127) Or boolSlamCollision)))
            If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, "          <<<<<  subDriveToDestinationLoop  START     ftnElapsedTime() = " & ftnElapsedTime())
            subCheckForPause()
            GotABotClass.subRedraw(tbpBotPose)

            If Not GotABotClass.InvokeRequired Then GotABotClass.StatusDriving.ForeColor = Color.Black
            Dim tbpPoseBeforMove As typeBotPose = tbpBotPose
            If boolDebugDrive Then frmDebug.WritePose(" subDriveToDestinationLoop: " & ftnElapsedTime() & " tbpBotPose = ", tbpBotPose)
            tbpTargetPose = ftnFindTargetPose(tbpBotPose) ' 
            If boolDebugDrive Then frmDebug.WritePose(" subDriveToDestinationLoop: " & ftnElapsedTime() & "  Target Pose = ", tbpTargetPose)
            tbpBotPose = ftnControlDriving(tbpBotPose, tbpTargetPose, strTypeOfMove)  '  
            tbpBotPose = ftnPostMovement(tbpBotPose, strTypeOfMove, tbpPoseBeforMove, tbpTargetPoseGlobal)
            subDoEventsIfMainThread()
            If (ftnPoseUncertaintyFactor(tbpBotPose) < 288) Then  '(288 = (12 inches + 12 inches) * 12 degrees) _
                boolPoseConfirmedAfterCollision = True
            End If
            If boolSlam Then
                If Not ftnGoodLocalization() Then subScanForLandMarks()
                boolGoodPath = ftnFindSlamPath()
            End If
            If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, "          <<<<<  subDriveToDestinationLoop  END     ftnElapsedTime() = " & ftnElapsedTime())
            strTypeOfMove = ""
        Loop  ' 
    End Sub

    Sub subSpeakDestinationSuccess(ByVal tmcDestination As typeMapCoordinate)
        If Not boolSlam And Not boolDock Then
            If (ftnDistanceBetweenLocations(tbpBotPose.Location, tmcDestination) <= (sngBotDiameter / 6)) Then
                ftnSendCommand("play phrase 'nailed it.'")
            ElseIf (ftnDistanceBetweenLocations(tbpBotPose.Location, tmcDestination) <= sngBotDiameter / 3) Then
                ftnSendCommand("play phrase 'close enough'")
            End If
        End If
    End Sub

    Function ftnControlDriving(ByVal tbpEstimatedPose As typeBotPose, ByVal tbpTargetPose As typeBotPose, ByRef strTypeOfMove As String) As typeBotPose
        On Error Resume Next
        ftnControlDriving = tbpEstimatedPose
        If boolDebugDrive And Not boolBotIsMoving Then frmDebug.subWriteLogs(Color.Red, "          <<<<<  ftnControlDriving  START     ftnElapsedTime() = " & ftnElapsedTime())
        tbpTargetPoseGlobal = tbpTargetPose
        Dim sngTurnAngle As Single = ftnTurnAngle(tbpEstimatedPose, tbpTargetPose.Location)
        strTypeOfMove = ftnTypeOfMove(tbpEstimatedPose, tbpTargetPose)
        If boolStop Then
            strTypeOfMove = "Stop"
            ftnControlDriving = ftnBotStop(ftnControlDriving)
        ElseIf Not boolBotIsMoving And ((strTypeOfMove = "Fwd") Or (strTypeOfMove = "XY")) And ftnObstacleDetected(tbpEstimatedPose, strTypeOfMove, tbpTargetPoseGlobal) Then
            ftnControlDriving = ftnObstacleAvoid(tbpEstimatedPose, tbpTargetPoseGlobal)
        ElseIf Not boolStop Then
            Select Case strTypeOfMove
                Case "Turn"
                    ' TODO - move back up function to modNavigationPath 
                    Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tbpTargetPose.Location)
                    If (Abs(sngTurnAngle) < 135) Or (sngDistanceToTarget > sngTenInches) Then
                        tbpTargetPoseGlobal = tbpEstimatedPose
                        tbpTargetPoseGlobal.Orientation = ftnConvertTo360(ftnControlDriving.Orientation + sngTurnAngle)
                        ftnControlDriving = ftnBotTurn(ftnControlDriving, True, tbpTargetPoseGlobal)
                    Else
                        frmDebug.subWriteLogs(Color.Red, Now() & " ***** Turn and Backup: . ***************************")
                        tbpTargetPoseGlobal.Orientation = ftnConvertTo360(ftnControlDriving.Orientation + sngTurnAngle - 180)
                        ftnControlDriving = ftnBotTurn(ftnControlDriving, False, tbpTargetPoseGlobal)  ' prepare to backup
                        boolPoseConfirmedAfterCollision = False  ' force localization
                        ftnControlDriving = ftnBotFwd(ftnControlDriving, tbpTargetPoseGlobal) ' backup
                    End If
                Case "Fwd"
                    ftnControlDriving = ftnBotFwd(ftnControlDriving, tbpTargetPoseGlobal)
                Case "XY"
                    tbpTargetPoseGlobal.Orientation = ftnConvertTo360(ftnControlDriving.Orientation + sngTurnAngle)
                    ftnControlDriving = ftnBotXY(ftnControlDriving, tbpTargetPoseGlobal)
                Case "Stop"
                    ftnControlDriving = ftnBotStop(ftnControlDriving)
            End Select
        End If
        If boolDebugDrive And Not boolBotIsMoving Then frmDebug.subWriteLogs(Color.Red, "          <<<<<  ftnControlDriving  END     ftnElapsedTime() = " & ftnElapsedTime())
        subDoEventsIfMainThread()
    End Function



End Module
