﻿

' This is the top level interface to the Bot.  Commands are parsed, interpreted and sent to the specific Bot IF.
' COMMAND SYNTAX
'    Move 10 i
'    Move 90 degrees
'    battery
'    position
'    input analog 1
'    input digital
'    play 
'    bumper
'    set
'    stop
Imports System.Threading
Imports System

Imports System.Speech.synthesis
Imports System.Speech
Imports System.Speech.Recognition
Imports System.Math


Module modIF_BotTop
    Dim strRecv As String

    Public Function ftnSendCommand(ByVal strCommand As String) As String
        Dim strParsedText As String() = strCommand.Split(" "c)
        ftnSendCommand = ""
        boolSend_Complete = False
        Try
            If (boolSimulatedMovement = False) Or (strParsedText(0) = "play") Then ' 
                Select Case strParsedText(0)
                    Case "move" : If Not boolStop Then _
                            ftnSendCommand = ftnMoveCommand(strParsedText, strCommand)
                    Case "battery" : ftnSendCommand = ftnGetBattery(strParsedText, strCommand)
                    Case "input" : If (UBound(strParsedText) > 0) Then
                            If (strParsedText(1) = "analog") Then  ' get  voltage
                                ftnSendCommand = ftnGetAnalog(strParsedText, strCommand)
                            ElseIf (strParsedText(1) = "digital") Then
                                ftnSendCommand = ftnGetDigital()
                            End If
                        End If
                    Case "play" : ftnSendCommand = ftnSpeak(strCommand)
                    Case "position" : ftnSendCommand = ftnGetPosition(strParsedText, strCommand) ' get odometry data
                    Case "bumper" : ftnSendCommand = ftnGetBumper(strParsedText, strCommand) ' get bumper switch status
                    Case "set" : ftnSendCommand = ftnSet(strParsedText, strCommand) ' Set bot parameter
                    Case "stop" : ftnSendCommand = ftnStop(strParsedText, strCommand) 'stop
                    Case Else : If (boolER1Selected) Then
                            ftnSendCommand = ftnER1_Communicate(strCommand, "OK", 1500)
                        ElseIf boolRoboRealmBotSelected Then
                            ftnSendCommand = ftnTransmitRR(strCommand, "", 0)
                        End If
                End Select
            End If
        Catch oEX As Exception
            subBotComError(strCommand, oEX)
        End Try
        boolSend_Complete = True
        If Not boolBotIsMoving Then subDoEventsIfMainThread()
        If boolSimulatedMovement Then subWait(1000)
    End Function


    Private Function ftnMoveCommand(ByVal strParsedText As String(), ByVal strCommand As String) As String
        On Error Resume Next
        ftnMoveCommand = "OK"
        Dim boolMidCourseCorrectionCommand As Boolean = boolBotIsMoving
        Dim sngMoveDistanceOrAngle As Single
        If (strTypeOfMove = "XY") Then
            sngMoveDistanceOrAngle = Sqrt(Convert.ToSingle(Val(strParsedText(2))) ^ 2 + Convert.ToSingle(Val(strParsedText(4))) ^ 2) / Cos(Atan(Convert.ToSingle(Val(strParsedText(4))) / Convert.ToSingle(Val(strParsedText(2)))))
        Else
            sngMoveDistanceOrAngle = Convert.ToSingle(Val(strParsedText(1)))
        End If
        Dim sngMovetimeOut As Single  ' Max time to wait for expected response

        If boolRoboRealmBotSelected Then
            If (boolEnableLocalizationWhileMoving Or boolEnableMidCourseCorrection) Then _
              subStartImageRecognition()
            timeMoveCommandSent = Now()
            ControlRRBot(sngMoveDistanceOrAngle, strTypeOfMove)
        ElseIf boolER1Selected Then
            If (strTypeOfMove = "Turn") Then
                sngMovetimeOut = Abs(sngMoveDistanceOrAngle / sngBotAngularVelocity * 2500) + sngStartOfMoveDelay + 4000
                If (sngMoveDistanceOrAngle > 0) Then
                    sngMoveDistanceOrAngle += constER1TurnHysterises
                Else
                    sngMoveDistanceOrAngle += -constER1TurnHysterises
                End If
                strCommand = "move " & sngMoveDistanceOrAngle & " " & strParsedText(2)
            ElseIf (strTypeOfMove = "Fwd") Then
                sngMovetimeOut = Abs(ftnMapUnitsToInches(sngMoveDistanceOrAngle) / (sngBotLinearVelocity * 2.54) * 2200) + sngStartOfMoveDelay + 4000
                If (sngMoveDistanceOrAngle > 0) Then
                    sngMoveDistanceOrAngle += ftnInchesToMapUnits(constER1FwdHysterises)
                Else
                    sngMoveDistanceOrAngle -= ftnInchesToMapUnits(constER1FwdHysterises * 0.33333)
                End If
                strCommand = "move " & sngMoveDistanceOrAngle & " " & strParsedText(2)
            ElseIf (strTypeOfMove = "XY") Then
                sngMovetimeOut = Abs(ftnMapUnitsToInches(sngMoveDistanceOrAngle) / (ftnMinValue(sngBotLinearVelocity, 30) * 2.54) * 2200) + sngStartOfMoveDelay + 15000
            End If
            ftnER1_Communicate(strCommand, "OK", 1000)
            timeMoveCommandSent = Now()
            boolBotIsMoving = True
            If boolMidCourseCorrectionCommand Then
                ftnMoveCommand = ftnER1_Communicate("events", "", 0)
            Else
                If (boolEnableLocalizationWhileMoving Or boolEnableMidCourseCorrection) Then subStartImageRecognition()
                ftnMoveCommand = ftnER1_Communicate("events", " done", sngMovetimeOut)
            End If
        End If
    End Function


    Private Function ftnGetPosition(ByVal strParsedText As String(), ByVal strCommand As String) As String
        'Return position as x,y in map units
        ' TODO Add RR Code
        On Error Resume Next
        ftnGetPosition = "ERROR :"
        Dim strResponse As String
        If boolRoboRealmBotSelected Then
            ftnGetPosition = "OK"
            ftnGetPosition = ftnGetPosition & tbpBotPose.Location.X.ToString   ' TODO tbpBotPose may be stale
            ftnGetPosition = ftnGetPosition & " " & tbpBotPose.Location.Y.ToString
        ElseIf boolER1Selected Then
            strResponse = ftnER1_Communicate("position", vbCr, 1000)
            strParsedText = strResponse.Split(" "c)
            If (UBound(strParsedText) = 5) Then
                If (strParsedText(0) = "OK") Then
                    ftnGetPosition = "OK "
                    ftnGetPosition &= ftnCmToMapUnits(Val(strParsedText(1))).ToString
                    ftnGetPosition &= " " & ftnCmToMapUnits(Val(strParsedText(2))).ToString
                Else
                    ftnGetPosition &= strResponse
                End If
            Else
                ftnGetPosition &= strResponse
            End If
            End If
    End Function


    Private Function ftnGetBumper(ByVal strParsedText As String(), ByVal strCommand As String) As String
        'Return position as x,y in map units
        ' TODO Add RR Code
        On Error Resume Next
        ftnGetBumper = ""
        Dim strResponse As String
        If boolRoboRealmBotSelected Then
        ElseIf boolER1Selected Then
            Dim byteGetDigital As Byte = ftnGetDigital() And 7
            If (byteGetDigital = 0) Then
                ftnGetBumper = "7"  '  probably a COM error, ignore
            Else
                ftnGetBumper = byteGetDigital.ToString
            End If
        End If
    End Function



    Private Function ftnStop(ByVal strParsedText As String(), ByVal strCommand As String) As String
        ' TODO Add RR Code
        On Error Resume Next
        ftnStop = ""
        If boolRoboRealmBotSelected Then
            ControlRRBot(0, "Fwd")
        ElseIf boolER1Selected Then
            ftnStop = ftnER1_Communicate(strCommand, "OK", 1000)
        End If
    End Function


    Private Function ftnSet(ByVal strParsedText As String(), ByVal strCommand As String) As String
        ' TODO Add RR Code
        On Error Resume Next
        ftnSet = ""
        If boolRoboRealmBotSelected Then
        ElseIf boolER1Selected Then
            ftnSet = ftnER1_Communicate(strCommand, "OK", 1000)
        End If
    End Function


    Private Function ftnGetBattery(ByVal strParsedText As String(), ByVal strCommand As String) As String
        On Error Resume Next
        Dim intBatteryPercentage As Integer = 0
        If boolRoboRealmBotSelected Then
            sngBotBatteryVoltage = CSng(Val(ftnGetVariableFromRR("battery"))) / 127 * 12.5
            intBatteryPercentage = Int((100.0 * (sngBotBatteryVoltage - 11.35) / 0.75))
            If (intBatteryPercentage > 100) Then intBatteryPercentage = 100
            If (intBatteryPercentage < 0) Then intBatteryPercentage = 0
            intBatteryPercentage = Int(intBatteryPercentage / 20) * 20
            If Not GotABotClass.InvokeRequired Then strStatusBotVoltage = "Bot: " & intBatteryPercentage.ToString & "%"

        ElseIf boolER1Selected And Not boolBotIsMoving Then
            strRecv = ftnER1_Communicate("input analog 15", "OK", 1100)
            strParsedText = strRecv.Split(" "c)
            If (UBound(strParsedText) >= 1) And (strParsedText(0) = "OK") Then
                If (Val(strParsedText(1)) > 0) Then
                    sngBotBatteryVoltage = Convert.ToSingle(Val(strParsedText(1)) / 65536.0 * 16.4)
                End If

                strStatusBotVoltage = "Bot: " & sngBotBatteryVoltage.ToString("N1") & "V"
            ElseIf (strParsedText(0) = "Error:") Or (sngBotBatteryVoltage = 0) Then
                sngBotBatteryVoltage = 0
                strStatusBotVoltage = "   ??   "
            End If
        End If
        ftnGetBattery = strStatusBotVoltage
        If (ftnGetBattery = "   ??   ") Then ftnSendCommand("play phrase 'No power'")
        sngLaptopBatteryPercent = Int(SystemInformation.PowerStatus.BatteryLifePercent * 100)
        GotABotClass.subUpdateAllText(tbpBotPose)
    End Function


    Private Function ftnGetAnalog(ByVal strParsedText As String(), ByVal strCommand As String) As String
        On Error Resume Next
        intAnalogReading = 0
        ftnGetAnalog = ""
        If boolER1Selected Then
            strRecv = ftnER1_Communicate(strCommand, "OK", 1500) '.1
            strParsedText = strRecv.Split(" "c)
            If (strParsedText(0) = "OK") Then
                If (UBound(strParsedText) >= 1) Then
                    intAnalogReading = 0
                    For I As Integer = 1 To UBound(strParsedText)
                        If (strParsedText(I) IsNot Nothing) And (strParsedText(I) <> "") Then
                            intAnalogReading = Convert.ToInt32(Convert.ToDouble(strParsedText(I)))
                        End If
                        ftnGetAnalog &= intAnalogReading.ToString & " "
                    Next I
                End If
            End If
        End If
    End Function

    Public strGetDigital As String = ""
    Public Function ftnGetDigital() As Byte
        On Error Resume Next
        Dim strParsedText As String()
        ftnGetDigital = "&Hff"
        If boolER1Selected Then
            strRecv = ftnER1_Communicate("input digital", "OK f", 1500)  '.1
            strGetDigital = strRecv
            strParsedText = strRecv.Split(" "c)
            If (UBound(strParsedText) >= 1) Then
                If (strParsedText(0) = "OK") Then _
                    ftnGetDigital = Val("&H" & strParsedText(1))
            End If
        End If
    End Function

    Private Function ftnSpeak(ByVal strCommand As String) As String
        On Error Resume Next
        Dim strParsedText As String() = strCommand.Split(" "c)
        If boolER1Selected And Not boolMute And Not boolPCVoice Then
            If boolMute Then
                ftnSpeak = "OK"
            Else
                Thread.Sleep(100)
                subDoEventsIfMainThread()
                ftnSpeak = ftnER1_Communicate(strCommand, "OK", 1000)
                ftnSpeak &= ftnER1_Communicate("events", "play done", 1500 + UBound(strParsedText) * 650)
            End If
        Else
            If Not boolMute Then
                Dim synth As New Synthesis.SpeechSynthesizer
                strCommand = strCommand.Replace("play phrase", "")
                strCommand = strCommand.Replace("'", "")
                synth.Speak(strCommand)
            End If
            ftnSpeak = "OK"
            'subWait(UBound(strParsedText) * 500)
        End If
    End Function

    Private Sub subBotComError(ByVal strCommand As String, ByVal oEX As Exception)
        On Error Resume Next
        frmDebug.subWriteLogs(Color.Red, " SocketException While Sending: " & strCommand & ".    Error Message Received: " & oEX.Message)
        frmDebug.subWriteLogs(Color.Red, "  ")
        frmDebug.subWriteLogs(Color.Red, "  ----------------Stack Trace-------------------")
        frmDebug.subWriteLogs(Color.Red, System.Environment.StackTrace)
        frmDebug.subWriteLogs(Color.Red, "  ")
        If Not boolSlam Then GotABotClass.subRedraw(tbpBotPose)
        If boolER1Selected Then
            Thread.Sleep(2000)
            frmDebug.subWriteLogs(Color.Red, "   Driving terminated due to Bot communication error.")
            'ER1Disconnect()
        End If
    End Sub

End Module
