﻿
' Interface to a Local Evolution RCC for image recognition

Imports System
Imports System.Net
Imports System.Net.Sockets
Imports System.Text
Imports System.Threading
Imports Microsoft.VisualBasic
Imports System.EventArgs



Module modLocalRccClient
    Public strLocalRccSendMessage As String = ""
    Public strLocalRccReceiveMessage As String = ""
    Public evntLocalRccDataWaitingToSend As New ManualResetEvent(False)
    Public boolLocalRccDataWaitingToSend As Boolean = False
    Public evntLocalRccDataSent As New ManualResetEvent(False)
    Public evntLocalRccDataReceived As New ManualResetEvent(False)
    Public evntLocalRccClientDataPassed As New ManualResetEvent(False)
    Dim intNoResponseCounter As Integer = 0
    Dim boolLRccLoggedIn As Boolean = False

    Public Function ftnLRCC_Communicate(ByVal strCommand As String, ByVal strResponse As String, ByVal inttimeOut As Integer) As String
        On Error Resume Next
        ftnLRCC_Communicate = ""
        If boolEnableLocalRCC And classLRCCClient.Connected And Not boolShutDown Then
            If Not boolLRccLoggedIn Then ftnLogin()
            ftnSendCommandLRCC(strCommand)
            If (strResponse <> "") Then ftnLRCC_Communicate = ftnWaitForExpectedResponseLRCC(strResponse, inttimeOut)
            If Not boolLRccLoggedIn Or CBool(InStr(ftnLRCC_Communicate, "login")) Then ftnLogin()
        End If
    End Function

    Function ftnLogin() As String
        boolLRccLoggedIn = True
        ftnLogin = ftnLRCC_Communicate("Login " & strPassword, "OK", 1000)
        boolLRccLoggedIn = Not CBool(InStr(ftnLogin, "login"))
        If boolLRccLoggedIn And Not boolShutDown Then
            ftnLRCC_Communicate("sense speech off", "OK", 1000)
            ftnLRCC_Communicate("set collision detection off", "OK", 1000)
            ftnLRCC_Communicate("set obstacle avoidance off", "OK", 1000)
            ftnLogin = ftnLRCC_Communicate("sense objects on", "OK", 1000)
        End If
    End Function

    Function ftnSendCommandLRCC(ByVal strCommand As String) As String
        ftnSendCommandLRCC = ""
        If boolEnableLocalRCC And classLRCCClient.Connected And Not boolShutDown Then
            Try
                evntLocalRccClientDataPassed.Reset()
                strLocalRccSendMessage = strCommand & vbCrLf
                boolLocalRccDataWaitingToSend = True
                evntLocalRccDataWaitingToSend.Set()
                'GC.Collect()
            Catch e As Exception
                ftnSendCommandLRCC = ""
            End Try
        End If
    End Function


    Function ftnWaitForExpectedResponseLRCC(ByVal strTextToWaitFor As String, ByVal inttimeOut As Integer) As String
        On Error Resume Next
        Dim timeOut As DateTime = Now().AddMilliseconds(inttimeOut)
        ftnWaitForExpectedResponseLRCC = ""
        Do
            ftnWaitForExpectedResponseLRCC &= ftnGetTextLRCC(150)
        Loop Until ((Now() > timeOut) Or CBool(InStr(ftnWaitForExpectedResponseLRCC, strTextToWaitFor)) Or boolStop Or boolShutDown)
    End Function

    Dim boolWaitingForData As Boolean = False
    Public Function ftnGetTextLRCC(ByVal intTimeout As Integer) As String  'must be thread safe
        On Error Resume Next
        ftnGetTextLRCC = ""
        If classLRCCClient.Connected And Not boolShutDown Then
            boolWaitingForData = True
            If (intTimeout > 0) Then evntLocalRccDataReceived.WaitOne(intTimeout) ' 
            ftnGetTextLRCC = strLocalRccReceiveMessage
            strLocalRccReceiveMessage = ""
            boolWaitingForData = False
            evntLocalRccDataReceived.Reset()
            evntLocalRccClientDataPassed.Set()
        End If
    End Function



#Region "LocalRccClient"

    '+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    '
    '
    '  Example adapted from http://msdn.microsoft.com/en-us/library/bew39x2a(v=VS.100).aspx
    '

    '
    '+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    Dim boolErrorDetected As Boolean = False

    Public Class classLRCCClient
        Public Shared evntConnectDone As New ManualResetEvent(False) ' ManualResetEvent instances signal completion.
        Public Shared Connected As Boolean = False



        Public Sub subMain()
            On Error Resume Next
            While Not boolShutDown
                classLRCCClient.Connected = False
                boolErrorDetected = False
                intNoResponseCounter = 0
                Dim LocalRCCEndPoint As New IPEndPoint(ipAddress.Parse("127.0.0.1"), intRobot_Port + 200) ' Create a TCP/IP socket.
                Dim LocalRCCClientSocket As New Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp)
                LocalRCCClientSocket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReceiveTimeout, 1500)
                LocalRCCClientSocket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.SendTimeout, 1500)
                LocalRCCClientSocket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, 1)
                LocalRCCClientSocket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.DontLinger, True)
                If boolEnableLocalRCC And Not classLRCCClient.Connected And Not boolErrorDetected And Not boolShutDown Then 'Connect and initialize
                    Thread.Sleep(100)
                    evntConnectDone.Reset()
                    LocalRCCClientSocket.BeginConnect(LocalRCCEndPoint, New AsyncCallback(AddressOf subConnectCallback), LocalRCCClientSocket)
                    evntConnectDone.WaitOne(1500)
                    If classLRCCClient.Connected And Not boolShutDown Then subReceive(LocalRCCClientSocket)
                End If

                intNoResponseCounter = 0
                While classLRCCClient.Connected And Not boolErrorDetected And Not boolShutDown ' subSend commands 
                    evntLocalRCCDataWaitingToSend.WaitOne(1000)
                    evntLocalRccDataWaitingToSend.Reset()
                    If boolLocalRccDataWaitingToSend And Not boolErrorDetected And Not boolShutDown Then
                        boolLocalRccDataWaitingToSend = False
                        subSend(LocalRCCClientSocket, strLocalRccSendMessage)
                        If (intNoResponseCounter > 2) Then boolErrorDetected = True
                    End If
                    Thread.Sleep(10)
                    'GC.Collect()
                End While
                subDisconnect(LocalRCCClientSocket)
            End While
        End Sub 'subMain

        Private Sub subDisconnect(ByRef client As Socket)
            Try
                classLRCCClient.Connected = False
                intNoResponseCounter = 0
                Thread.Sleep(500)
                If (client IsNot Nothing) Then
                    If client.Connected Then
                        client.Shutdown(SocketShutdown.Both)
                        Thread.Sleep(500)
                        If client.Connected Then client.Disconnect(Not boolShutDown)
                        Thread.Sleep(500)
                    End If
                    client.Dispose()
                End If
                Thread.Sleep(500)
            Catch ex As Exception
                classLRCCClient.Connected = False
                boolErrorDetected = True
            End Try
        End Sub

        Private Shared Sub subConnectCallback(ByVal ar As IAsyncResult)
            Try
                If Not boolErrorDetected And Not boolShutDown And (ar IsNot Nothing) Then
                    Dim client As Socket = CType(ar.AsyncState, Socket) ' Retrieve the socket from the LocalRCCClientState object.
                    If (client IsNot Nothing) Then
                        If client.Connected And Not boolShutDown Then
                            client.EndConnect(ar) ' Complete the connection.
                            evntConnectDone.Set() ' Signal that the connection has been made.
                        End If
                        classLRCCClient.Connected = client.Connected
                    End If
                End If
            Catch e As Exception
                boolErrorDetected = True
                classLRCCClient.Connected = False
            End Try
        End Sub 'subConnectCallback


        Private Shared Sub subReceive(ByVal client As Socket)
            If classLRCCClient.Connected And Not boolShutDown And Not boolErrorDetected And (client IsNot Nothing) Then
                Try
                    Dim state As New typeSocketClass ' Create the state object
                    state.workSocket = client
                    ' Begin receiving the data from the remote device.
                    client.BeginReceive(state.byteReceiveBuffer, 0, typeSocketClass.BufferSize, 0, New AsyncCallback(AddressOf subReceiveCallback), state)
                    'GC.Collect()
                Catch e As Exception
                    boolErrorDetected = True
                End Try
            End If
        End Sub 'subReceive


        Private Shared Sub subReceiveCallback(ByVal ar As IAsyncResult)
            Try
                If classLRCCClient.Connected And Not boolShutDown And Not boolErrorDetected And (ar IsNot Nothing) Then
                    Dim LocalRCCSocketClassInstance As typeSocketClass = CType(ar.AsyncState, typeSocketClass)
                    Dim LocalRCCClientSocket As Socket = LocalRCCSocketClassInstance.workSocket
                    Dim bytesRead As Integer = LocalRCCClientSocket.EndReceive(ar) ' Read data from the client socket.
                    If classLRCCClient.Connected And Not boolErrorDetected And bytesRead > 0 And Not boolShutDown _
                        And (LocalRCCClientSocket IsNot Nothing) Then ' There  might be more data, so store the data received so far.
                        If (LocalRCCSocketClassInstance.sbReceivedData.Length < (UBound(LocalRCCSocketClassInstance.byteReceiveBuffer) - bytesRead)) Then
                            LocalRCCSocketClassInstance.sbReceivedData.Append(Encoding.ASCII.GetString(LocalRCCSocketClassInstance.byteReceiveBuffer, 0, bytesRead) & " ")
                            LocalRCCSocketClassInstance.sbReceivedData.Append(" ")
                            If (strLocalRccReceiveMessage.Length + bytesRead) < UBound(LocalRCCSocketClassInstance.byteReceiveBuffer) Then
                                strLocalRccReceiveMessage &= LocalRCCSocketClassInstance.sbReceivedData.ToString() ' Throws exceptions - out of range
                            End If
                        End If
                        If strLocalRccReceiveMessage.Length > 3 And boolWaitingForData Then
                            If CBool(InStr(strLocalRccReceiveMessage, vbCr)) Then
                                evntLocalRccDataReceived.Set() 'Signal the subMain thread to continue.
                                evntLocalRccClientDataPassed.WaitOne(500) ' subWait for subMain thread to get the data
                                evntLocalRccClientDataPassed.Reset()
                                LocalRCCSocketClassInstance.sbReceivedData.Clear()   'clear the buffer
                                strLocalRccReceiveMessage = ""
                            End If
                        End If
                    End If
                    If classLRCCClient.Connected And Not boolShutDown And Not boolErrorDetected And (LocalRCCClientSocket IsNot Nothing) Then _
                        LocalRCCClientSocket.BeginReceive(LocalRCCSocketClassInstance.byteReceiveBuffer, 0, typeSocketClass.BufferSize, 0, New AsyncCallback(AddressOf subReceiveCallback), LocalRCCSocketClassInstance)
                Else
                    boolErrorDetected = True
                End If 'If classLRCCClient.Connected
            Catch ex As Exception
                boolErrorDetected = True
            End Try
        End Sub 'subReadCallback

        Private Sub subSend(ByVal LocalRCCClientSocket As Socket, ByVal strMessage As String)
            ' Convert the string strMessage to byte strMessage using ASCII encoding.
            Dim bytestrMessage As Byte() = Encoding.ASCII.GetBytes(strMessage)
            ' Begin sending strMessage to LocalRCC.
            Try
                If classLRCCClient.Connected And Not boolErrorDetected And Not boolShutDown And (LocalRCCClientSocket IsNot Nothing) Then _
                  LocalRCCClientSocket.BeginSend(bytestrMessage, 0, bytestrMessage.Length, 0, New AsyncCallback(AddressOf subSendCallback), LocalRCCClientSocket)
                'GC.Collect()
            Catch ex As Exception
                boolErrorDetected = True
            End Try

            evntLocalRCCDataSent.WaitOne(300)
            evntLocalRCCDataSent.Reset()
        End Sub 'subSend


        Private Sub subSendCallback(ByVal ar As IAsyncResult)
            Try
                If classLRCCClient.Connected And Not boolErrorDetected And Not boolShutDown And (ar IsNot Nothing) Then
                    Dim LocalRCCClientSocket As Socket = CType(ar.AsyncState, Socket) ' Retrieve the socket from the LocalRCCSocketClassInstance object.
                    evntLocalRCCDataSent.Set()
                End If
            Catch ex As Exception
                evntLocalRCCDataSent.Set()
                boolErrorDetected = True
            End Try
        End Sub 'subSendCallback

    End Class 'LocalRccClientServerClass

#End Region




End Module
