﻿

' This module returns a bot pose based upon Range Scan Data.
' It is used for devices which can return a 2D range scan such as a Kinect or Laser Range scanner
Imports System
Imports System.Net
Imports System.Net.Sockets
Imports System.Text
Imports System.Threading
Imports Microsoft.VisualBasic
Imports System.EventArgs
Imports System.Math
Imports Microsoft.Kinect

Module modLocalize2Dscan
    Const sngWeakCorrelation As Single = 0.8
    Const sngGoodCorrelation As Single = 1.0
    Const sngExcellentCorrelation As Single = 1.35
    Dim bool2DScanLocalizationInProgress As Boolean = False ' dont want conflict between modLocalizeTop and subLocalizeWhileMoving2Dscan_Top

    Public Function ftnLocalizeFrom2DScan_Top(ByRef tbpEstimatedPose As typeBotPose, ByVal tbpReferencePose As typeBotPose, ByVal boolBotIsMoving As Boolean, _
                                              ByRef boolGood2DLocalization As Boolean, ByRef sngCorrelationFactor As Single) As typeBotPose
        On Error Resume Next
        Dim tbpRawScanPose As typeBotPose
        sngCorrelationFactor = 0
        If boolBotIsMoving Or ftnIAmLost(tbpEstimatedPose) Then
            intNumberOfScanVectors = 64 ' reduce number of vectors while moving or lost for faster localization
        Else
            intNumberOfScanVectors = 128
        End If
        boolGood2DLocalization = False
        boolAbortLocalization = False
        ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
        If (boolKinectConnected Or boolRangersAreEnabled) And Not boolStop And Not boolSimulatedMovement Then
            Dim strStatusDrivingRoot As String = strStatusDriving
            Dim boolGoodPreVal2DLocalization As Boolean = False
            If Not boolBotIsMoving Then subDoEventsIfMainThread()
            If bool2DScanLocalizationInProgress Then
                Dim timeOut As DateTime = Now().AddMilliseconds(12000)
                While bool2DScanLocalizationInProgress And Not boolStop   ' wait for previous scan to finish
                    boolAbortLocalization = True
                    subDoEventsIfMainThread()
                    Thread.Sleep(25)
                    If Now() >= timeOut Then
                        ftnLocalizeFrom2DScan_Top = ftnLocalizeUsingDeadReckoning(ftnLocalizeFrom2DScan_Top, constGotAKinectDelay, tbpTargetPoseGlobal)
                        frmDebug.subWriteLogs(Color.Green, "  ftnLocalizeFrom2DScan_Top timeOut ")
                        Exit Function
                    End If
                End While
            End If
            boolAbortLocalization = False
            bool2DScanLocalizationInProgress = True
            If boolDebugLocalize2D Then frmDebug.subWriteLogs(Color.Green, " Elapsed Time - pre localize:  " & ftnElapsedTime())
            tbpRawScanPose = ftn2dLocalize(tbpEstimatedPose, boolGoodPreVal2DLocalization, sngCorrelationFactor)
            If boolDebugLocalize2D Then frmDebug.subWriteLogs(Color.Green, " Elapsed Time - post localize:  " & ftnElapsedTime())
            If Not boolGoodPreVal2DLocalization Then
                ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
            Else
                ftnLocalizeFrom2DScan_Top = ftnFindAvgPose(tbpRawScanPose, tbpEstimatedPose, 0)
                ftnLocalizeFrom2DScan_Top = ftnMinPoseUncertainty(ftnLocalizeFrom2DScan_Top)
                If boolBotIsMoving Then tbpReferencePose = tbpEstimatedPose
                ftnLocalizeFrom2DScan_Top = ftnValidatePose(ftnLocalizeFrom2DScan_Top, tbpReferencePose, boolGood2DLocalization)
                If Not boolGood2DLocalization Then
                    If (sngCorrelationFactor > sngGoodCorrelation) And (ftnDistanceBetweenLocations(tbpRawScanPose.Location, tbpEstimatedPose.Location) > sngSixInches) Then ' Try without averaging poses
                        If boolDebugLocalize2D Then
                            frmDebug.subWriteLogs(Color.Green, " ================ Bad 2D Localization.  Recalculating without averaging. ")
                            ftnLocalizeFrom2DScan_Top = tbpRawScanPose
                            ftnLocalizeFrom2DScan_Top = ftnMinPoseUncertainty(ftnLocalizeFrom2DScan_Top)
                            ftnLocalizeFrom2DScan_Top = ftnValidatePose(ftnLocalizeFrom2DScan_Top, tbpReferencePose, boolGood2DLocalization)
                        End If
                    End If
                End If
            End If
            If Not boolBotIsMoving Then ' update graphics and log file with results
                strStatusDriving = strStatusDrivingRoot & " 2D=" & FormatNumber(sngCorrelationFactor, 2) & "; "
                GotABotClass.subRedraw(ftnLocalizeFrom2DScan_Top)
                If boolGood2DLocalization Then
                    subDrawScanVectors(tbpRawScanPose, True)
                    subWait(100)
                End If
                frmDebug.WritePose(ftnElapsedTime() & "       Pose Derived From 2D Scan Data  ", tbpRawScanPose)
                If boolGood2DLocalization Then
                    frmDebug.subWriteLogs(Color.Green, "                  Correlation Factor = " & FormatNumber(sngCorrelationFactor, 2))
                Else
                    frmDebug.subWriteLogs(Color.Red, "                    Correlation Factor = " & FormatNumber(sngCorrelationFactor, 2) & "   (2D Scan localization failed)")
                End If
            End If
        Else
        End If
        bool2DScanLocalizationInProgress = False
    End Function




    '--------------------------------------  Debug Graphics  ----------------------------------------------------------------------

    Public Sub subDrawScanVectors(ByVal tbpPose As typeBotPose, ByVal boolGreen As Boolean)
        On Error Resume Next
        If boolShowMore And Not boolBotIsMoving And Not GotABotClass.InvokeRequired Then
            Dim sngOrientaionOffset As Single = 29.25 + tbpPose.Orientation
            Dim sngOrientationFactor As Single = 58.5 / 128
            If boolKinectConnected Then
                Dim tmcVectorOrigin As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpPose, tmcScanVectorOffset)
                For intKinectScanVector As Short = 0 To 127 Step 2 ' Draw Scan Vectors in Green or Red
                    If (trvScanVectors(intKinectScanVector).sngRange > 0) Then _
                            subDrawVector(tmcVectorOrigin, trvScanVectors(intKinectScanVector).sngRange, _
                                          ftnConvertTo360(sngOrientaionOffset - sngOrientationFactor * intKinectScanVector), boolGreen) ' Show Scan Vector on map
                Next
            End If
            If boolRangersAreEnabled And boolGreen Then ' Draw Ranger Vectors in Green 
                Dim trvRangerVectors(intHighestEnabledRangerNumber) As typeRangeVector  ' Range vector - relative to Bot center
                For intRangerNumber As Short = 1 To intHighestEnabledRangerNumber ' 
                    If trfRangeFinders(intRangerNumber).boolEnable Then
                        trvRangerVectors(intRangerNumber) = frmRangeFinder.ftnRangeFinderVector(tbpPose, intRangerNumber, trfRangeFinders(intRangerNumber).sngRange, trfRangeFinders(intRangerNumber).intOrientation) ' Find Range vector - relative to Bot center
                        subDrawVector(tbpPose.Location, trvRangerVectors(intRangerNumber).sngRange, _
                            ftnConvertTo360(tbpPose.Orientation + trvRangerVectors(intRangerNumber).sngAngle), boolGreen)
                    End If
                Next
            End If
            subWait(100)
        End If
    End Sub

    Dim trvScanVectorsCopy(128) As typeRangeVector
    Dim boolftn2dLocalizeInProgress As Boolean = False
    Function ftn2dLocalize(ByVal tbpEstimatedPose As typeBotPose, _
                           ByRef boolGood2DLocalization As Boolean, ByRef sngCorrelationFactorBest As Single) As typeBotPose
        On Error Resume Next
        ftn2dLocalize = tbpEstimatedPose
        If Not boolftn2dLocalizeInProgress Then
            boolftn2dLocalizeInProgress = True
            sngCorrelationFactorBest = 0
            If Not boolBotIsMoving Then frmRangeFinder.subGetRangerReadings()
            If boolKinectConnected Or (boolRangersAreEnabled And Not boolBotIsMoving) Then
                If boolKinectConnected Then
                    'If boolDebugKinect Then frmDebug.subWriteLogs(Color.Black, " pre aquire Kinect Vectors:  " & ftnElapsedTime() & vbCrLf)
                    If evntVectorsAreFresh.WaitOne(1000) Then
                        If boolDebugKinect Then frmDebug.subWriteLogs(Color.Black, "  Kinect Vectors acquired by ftn2dLocalize:  " & ftnElapsedTime() & vbCrLf)
                        trvScanVectorsCopy = trvScanVectors
                        subDrawScanVectors(tbpEstimatedPose, False)
                    Else
                        frmDebug.subWriteLogs(Color.Green, "  TimeOut waiting for Kinect vectors")
                        Exit Function
                    End If
                End If
                If boolKinectConnected Then _
                  If ftnNonFlatWall() Then ftn2dLocalize = ftnPoseFrom2DScan(tbpEstimatedPose, sngCorrelationFactorBest)
            End If
            Dim boolGT5Inches As Boolean = (ftnDistanceBetweenLocations(tbpEstimatedPose.Location, ftn2dLocalize.Location) > sngFiveInches)
            If ftnIAmLost(tbpEstimatedPose) Then
                boolGood2DLocalization = (sngCorrelationFactorBest >= sngWeakCorrelation)
            ElseIf boolGT5Inches And (ftnPoseUncertaintyFactor(tbpEstimatedPose) > 288) Then  '(288 = (12 inches + 12 inches) * 12 degrees)
                boolGood2DLocalization = (sngCorrelationFactorBest >= sngExcellentCorrelation)
            ElseIf boolGT5Inches Then
                boolGood2DLocalization = (sngCorrelationFactorBest >= sngGoodCorrelation)
            Else
                boolGood2DLocalization = (sngCorrelationFactorBest >= sngWeakCorrelation)
            End If
        End If
            boolftn2dLocalizeInProgress = False
    End Function




    Public Function ftnPoseFrom2DScan(ByVal tbpEstimatedPose As typeBotPose, ByRef sngCorrelationFactorBest As Single) As typeBotPose
        On Error Resume Next
        sngCorrelationFactorBest = 0
        ftnPoseFrom2DScan = ftnPoseWithBestScanToMapCorrelation(tbpEstimatedPose, sngCorrelationFactorBest, sngFourInches, 2)
        If boolBotIsMoving Then
            ftnPoseFrom2DScan.LocationUncertainty.X = sngFourInches / sngGridSize
        Else
            ftnPoseFrom2DScan.LocationUncertainty.X = sngSixInches / sngGridSize
        End If
        ftnPoseFrom2DScan.LocationUncertainty.Y = ftnPoseFrom2DScan.LocationUncertainty.X
        ftnPoseFrom2DScan.OrientationUncertainty = 4
        ftnPoseFrom2DScan = ftnPoseWithBestScanToMapCorrelation(ftnPoseFrom2DScan, sngCorrelationFactorBest, sngOneInch, 1)
    End Function


    Dim boolAbortLocalization As Boolean = False
    Function ftnPoseWithBestScanToMapCorrelation(ByVal tbpEstimatedPose As typeBotPose, ByRef sngCorrelationFactorBest As Single, _
                                                 ByVal intLocationStep As Short, ByVal intOrientationStep As Short) As typeBotPose
        On Error Resume Next ' Find the pose within the bot's uncertainty region that most closely resembles the range data
        Dim intOrientationStart, intOrientationEnd As Short
        Dim tbpPossiblePose As typeBotPose = tbpEstimatedPose
        Dim intStartX, intEndX, intStartY, intEndY As Short
        Dim sngStdDeviation, sngStdDeviationBest, sngCorrelationFactor, sngCorrelationFactorQuick, sngCorrelationFactorQuickBest As Single
        Dim boolValidLocation As Boolean
        Dim boolSkipNextCheck As Boolean = False
        Dim intValidVectorCounter As Short = 0
        Dim intValidVectorCounterBest As Short = 0
        ftnPoseWithBestScanToMapCorrelation = tbpEstimatedPose
        intStartX = tbpEstimatedPose.Location.X - tbpEstimatedPose.LocationUncertainty.X
        If intStartX < LBound(byteNavigationMap, 1) Then intStartX = CShort(LBound(byteNavigationMap, 1))
        intEndX = tbpEstimatedPose.Location.X + tbpEstimatedPose.LocationUncertainty.X
        If intEndX >= UBound(byteNavigationMap, 1) Then intEndX = CShort(UBound(byteNavigationMap, 1)) - 1
        intStartY = tbpEstimatedPose.Location.Y - tbpEstimatedPose.LocationUncertainty.Y
        If intStartY < LBound(byteNavigationMap, 2) Then intStartY = CShort(LBound(byteNavigationMap, 2))
        intEndY = tbpEstimatedPose.Location.Y + tbpEstimatedPose.LocationUncertainty.Y
        If intEndY >= UBound(byteNavigationMap, 2) Then intEndY = CShort(UBound(byteNavigationMap, 2)) - 1
        subFindStartEndAngles(tbpEstimatedPose, intOrientationStart, intOrientationEnd)
        sngCorrelationFactorBest = ftnCorrelationFactor(tbpEstimatedPose, tbpEstimatedPose, sngStdDeviation, False, intValidVectorCounter) ' Find correlation at presumed bot location
        sngStdDeviationBest = sngStdDeviation
        sngCorrelationFactorQuickBest = ftnCorrelationFactor(tbpEstimatedPose, tbpEstimatedPose, sngStdDeviation, True, intValidVectorCounter)
        intValidVectorCounterBest = intValidVectorCounter
        If boolDebugLocalize2D Then frmDebug.subWriteLogs(Color.Green, " sngCorrelationFactorBest = " & sngCorrelationFactorBest)
        For tbpPossiblePose.Location.X = intStartX To intEndX Step intLocationStep ' look for better correlation throughout uncertainty region
            If boolStop Or boolAbortLocalization Then
                ftnPoseWithBestScanToMapCorrelation = tbpEstimatedPose
                sngCorrelationFactorBest = 0
                Exit Function
            End If
            For tbpPossiblePose.Location.Y = intStartY To intEndY Step intLocationStep
                If Not ftnInWall(tbpPossiblePose.Location) Then
                    For tbpPossiblePose.Orientation = intOrientationStart To intOrientationEnd Step intOrientationStep
                        If boolSkipNextCheck Then
                            boolSkipNextCheck = False
                        Else
                            tbpPossiblePose.Orientation = ftnConvertTo360(tbpPossiblePose.Orientation)
                            sngCorrelationFactorQuick = ftnCorrelationFactor(tbpEstimatedPose, tbpPossiblePose, sngStdDeviation, True, intValidVectorCounter) ' Find quick correlation at presumed bot location
                            'If boolDebugLocalize2D Then frmDebug.subWriteLogs(Color.Green, intLocationStep & "  " & tbpPossiblePose.Location.X & ",  " & tbpPossiblePose.Location.Y & " at " & tbpPossiblePose.Orientation & " = " & FormatNumber(sngCorrelationFactor, 2))
                            If (sngCorrelationFactorQuick < ftnMaxValue(sngCorrelationFactorQuickBest * 0.5, 0.4)) Then
                                boolSkipNextCheck = True  ' Skip the next check
                                'If boolDebugLocalize2D Then frmDebug.subWriteLogs(Color.Green, " SKIP")
                            ElseIf (sngCorrelationFactorQuick > sngCorrelationFactorQuickBest) Then
                                sngCorrelationFactor = ftnCorrelationFactor(tbpEstimatedPose, tbpPossiblePose, sngStdDeviation, False, intValidVectorCounter) ' Find correlation at tbpPossiblePose  location
                                If (sngCorrelationFactor > sngCorrelationFactorBest) Then
                                    If boolDebugLocalize2D Then
                                        frmDebug.subWriteLogs(Color.Green, " ================ New sngCorrelationFactorBest = " & FormatNumber(sngCorrelationFactor, 3) & "==============================")
                                        frmDebug.subWriteLogs(Color.Green, "Best Correlated Pose =  " & tbpPossiblePose.Location.X & ",  " & tbpPossiblePose.Location.Y & " at " & tbpPossiblePose.Orientation & " degrees")
                                        frmDebug.subWriteLogs(Color.Green, " intLocationStep = " & intLocationStep)
                                        frmDebug.subWriteLogs(Color.Green, " intValidVectorCounter = " & intValidVectorCounter)
                                        frmDebug.subWriteLogs(Color.Green, " sngStdDeviation = " & sngStdDeviation)
                                    End If
                                    ftnPoseWithBestScanToMapCorrelation = tbpPossiblePose
                                    sngCorrelationFactorQuickBest = sngCorrelationFactorQuick
                                    sngCorrelationFactorBest = sngCorrelationFactor
                                    sngStdDeviationBest = sngStdDeviation
                                    intValidVectorCounterBest = intValidVectorCounter
                                Else
                                End If
                            End If
                        End If
                    Next tbpPossiblePose.Orientation
                End If
            Next tbpPossiblePose.Location.Y
        Next tbpPossiblePose.Location.X
        ftnPoseWithBestScanToMapCorrelation.Orientation = ftnConvertTo360(ftnPoseWithBestScanToMapCorrelation.Orientation)
        ftnPoseWithBestScanToMapCorrelation.OrientationUncertainty = tbpEstimatedPose.OrientationUncertainty
        ftnPoseWithBestScanToMapCorrelation.LocationUncertainty = tbpEstimatedPose.LocationUncertainty
        ftnPoseWithBestScanToMapCorrelation = ftnPoseUncertaintyFrom2DScan(ftnPoseWithBestScanToMapCorrelation, sngCorrelationFactorBest, sngStdDeviationBest)
    End Function

    Function ftnCorrelationFactor(ByVal tbpEstimatedPose As typeBotPose, ByVal tbpPossiblePose As typeBotPose, ByRef sngStdDeviation As Single, _
                                      ByVal boolQuickCheck As Boolean, ByRef intValidVectorCounter As Short) As Single
        On Error Resume Next
        Dim int2DValidVectors As Short = 0
        Dim sng2DCorrelation As Single = 0
        intValidVectorCounter = 0
        sng2DCorrelation = ftnScanToMapCorrelation(tbpPossiblePose, sngStdDeviation, int2DValidVectors, boolQuickCheck)
        If Not boolBotIsMoving And boolRangersAreEnabled Then
            Dim trvRangerVectors(intHighestEnabledRangerNumber) As typeRangeVector  ' Range vector - relative to Bot center
            trvRangerVectors = ftnRangerVectorsRelativeBotCenter(tbpPossiblePose)
            ftnCorrelationFactor = ftnRangerToMapCorrelation(tbpPossiblePose, sngStdDeviation, intValidVectorCounter, boolQuickCheck, trvRangerVectors) ' 
            If boolKinectConnected Then
                ftnCorrelationFactor = sng2DCorrelation + ftnCorrelationFactor * intValidVectorCounter / int2DValidVectors
                intValidVectorCounter += int2DValidVectors
            End If
        Else
            ftnCorrelationFactor = sng2DCorrelation
            intValidVectorCounter = int2DValidVectors
        End If
    End Function

    Function ftnRangerVectorsRelativeBotCenter(ByVal tbpEstimatedPose As typeBotPose) As typeRangeVector()
        On Error Resume Next
        Dim trvRangerVector(intHighestEnabledRangerNumber) As typeRangeVector
        If boolRangersAreEnabled Then
            For intRangerNumber As Short = 1 To intHighestEnabledRangerNumber ' 
                If trfRangeFinders(intRangerNumber).boolEnable Then
                    trvRangerVector(intRangerNumber) = frmRangeFinder.ftnRangeFinderVector(tbpEstimatedPose, intRangerNumber, trfRangeFinders(intRangerNumber).sngRange, trfRangeFinders(intRangerNumber).intOrientation) ' Find Range vector - relative to Bot center
                End If
            Next intRangerNumber
        End If
        Return trvRangerVector
    End Function




    Function ftnScanToMapCorrelation(ByVal tbpPossiblePose As typeBotPose, ByRef sngStdDeviation As Single, _
                                     ByRef intValidVectorCounter As Short, ByVal boolQuickCheck As Boolean) As Single
        ' Find the correlation of a particular pose to the range data (higher =  better)
        ' boolQuickCheck = Only check for walls within +/-sngOneDotFiveSigma distance from the end of the vector
        On Error Resume Next
        ftnScanToMapCorrelation = 0
        sngStdDeviation = 0
        If boolKinectConnected Then
            Dim tmcScanPredictedWallLocation As typeMapCoordinate
            Dim boolGreySpace As Boolean = False
            Dim trvVector As typeRangeVector
            Dim sngDistanceError As Single
            Dim boolMaxRange As Boolean
            intValidVectorCounter = 0
            Dim tmcVectorOrigin As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpPossiblePose, tmcScanVectorOffset)
            Dim intSamplesPerVector As Short = 128 / intNumberOfScanVectors
            For intKinectScanVector As Short = 0 To 127 Step intSamplesPerVector
                'If (trvScanVectorsCopy(intKinectScanVector).sngRange = 0) Then
                '    trvScanVectorsCopy(intKinectScanVector) = trvScanVectorsCopy(intKinectScanVector + 1)
                'End If
                If (trvScanVectorsCopy(intKinectScanVector).sngRange = 0) Then  ' 
                    trvVector.sngRange = ftnInchesToMapUnits(intScanVectorsMinRange) * 0.5
                    trvVector.sngAngle = ftnConvertTo360(trvScanVectorsCopy(intKinectScanVector).sngAngle + tbpPossiblePose.Orientation)
                    sngDistanceError = ftnRangeVectorVsWallError(tmcVectorOrigin, trvVector, False)
                    If (Abs(sngDistanceError) < trvVector.sngRange) Then
                        sngStdDeviation += 1
                    Else
                        sngStdDeviation += 1.5  ' maybe no echo because of oblique angle or shiny object or ?
                    End If
                    intValidVectorCounter += 2
                Else
                    trvVector.sngRange = trvScanVectorsCopy(intKinectScanVector).sngRange
                    trvVector.sngAngle = ftnConvertTo360(trvScanVectorsCopy(intKinectScanVector).sngAngle + tbpPossiblePose.Orientation)
                    Dim tmcScanVectorEnd As typeMapCoordinate = ftnLocationFromLocation(tmcVectorOrigin, trvVector.sngAngle, trvVector.sngRange)
                    boolGreySpace = ftnGreySpace(tmcScanVectorEnd)
                    sngDistanceError = Abs(ftnRangeVectorVsWallError(tmcVectorOrigin, trvVector, boolQuickCheck))
                    boolMaxRange = (trvVector.sngRange > ftnInchesToMapUnits(intScanVectorsMaxRange))  ' there were no walls detected up to the maximum range of the kinect sensor
                    If boolMaxRange And (sngDistanceError < sngTwoSigma) Then
                        sngStdDeviation += 1
                        intValidVectorCounter += 2
                    ElseIf (sngDistanceError > sngOneDotFiveSigma) Then
                        If boolGreySpace Then
                            intValidVectorCounter += 2 ' Grey zone = cluttered environment / may have false echos
                            sngStdDeviation += 2.25
                        Else
                            intValidVectorCounter += 1
                            sngStdDeviation += 2.25
                        End If
                    ElseIf boolGreySpace Then
                        intValidVectorCounter += 3 ' Grey zone = cluttered environment / may have false echos
                        sngStdDeviation += ftnMinValue((sngDistanceError / sngOneSigma) ^ 2, 2.25)
                    Else
                        intValidVectorCounter += 3
                        sngStdDeviation += ftnMinValue((sngDistanceError / sngOneSigma) ^ 2, 2.25)
                    End If
                End If
            Next intKinectScanVector
            If (intValidVectorCounter > (0.55 * intNumberOfScanVectors)) Then
                intValidVectorCounter = intValidVectorCounter / 3
                sngStdDeviation = Sqrt(sngStdDeviation / intValidVectorCounter)
                sngStdDeviation = ftnMaxValue(sngStdDeviation, 0.001)
                ftnScanToMapCorrelation = intValidVectorCounter / (sngStdDeviation * intNumberOfScanVectors)
            End If
        End If
    End Function

    Function ftnRangerToMapCorrelation(ByVal tbpEstimatedPose As typeBotPose, ByRef sngStdDeviation As Single, ByRef intValidVectorCounter As Short, _
                                       ByVal boolQuickCheck As Boolean, ByVal trvRangerVectors() As typeRangeVector) As Single
        ' Find the correlation of a particular pose to the range data (higher =  better)
        ' boolQuickCheck = Don't check for walls furthur than x distance from the end of the vector
        On Error Resume Next
        Dim trvRangeVectorFromBotCenter As typeRangeVector
        ftnRangerToMapCorrelation = 0
        If boolRangersAreEnabled Then
            sngStdDeviation = 0
            Dim intNumberOfRangers As Short = 0
            Dim boolGreySpace As Boolean = False
            Dim sngDistanceError As Single
            Dim boolMaxRange As Boolean
            Dim trvVector As typeRangeVector
            intValidVectorCounter = 0
            For intRangerNumber As Short = 1 To intHighestEnabledRangerNumber ' 
                If trfRangeFinders(intRangerNumber).boolEnable Then
                    intNumberOfRangers += 1
                    boolMaxRange = (trfRangeFinders(intRangerNumber).sngRange > ftnCmToMapUnits(trfRangeFinders(intRangerNumber).sngMaxRange - 0.1))  ' there were no walls detected up to the maximum range of the ranger
                    If Not boolMaxRange Then
                        trvVector.sngRange = trvRangerVectors(intRangerNumber).sngRange
                        trvVector.sngAngle = ftnConvertTo360(trvRangerVectors(intRangerNumber).sngAngle + tbpEstimatedPose.Orientation)
                        Dim tmcRF_VectorEnd As typeMapCoordinate = ftnLocationFromLocation(tbpEstimatedPose.Location, trvVector.sngAngle, trvVector.sngRange)
                        boolGreySpace = ftnGreySpace(tmcRF_VectorEnd)
                        sngDistanceError = Abs(ftnRangeVectorVsWallError(tbpEstimatedPose.Location, trvVector, boolQuickCheck))
                    End If
                    If boolMaxRange Then
                        sngStdDeviation += 1
                        intValidVectorCounter += 2
                    ElseIf (sngDistanceError > sngOneSigma) Then
                        If boolGreySpace Then
                            intValidVectorCounter += 2 ' Grey zone = cluttered environment / may have false echos
                            sngStdDeviation += ftnMinValue((sngDistanceError / sngOneSigma) ^ 2, 1.5)
                        Else
                            intValidVectorCounter += 1 ' 
                            sngStdDeviation += ftnMinValue((sngDistanceError / sngOneSigma) ^ 2, 2.25)
                        End If
                    ElseIf boolGreySpace Then
                        intValidVectorCounter += 3 ' Grey zone = cluttered environment / may have false echos
                        sngStdDeviation += ftnMinValue((sngDistanceError / sngOneSigma) ^ 2, 1.5)
                    Else
                        intValidVectorCounter += 3
                        sngStdDeviation += ftnMinValue((sngDistanceError / sngOneSigma) ^ 2, 2.25)
                    End If
                End If
            Next intRangerNumber
            intValidVectorCounter = (intValidVectorCounter + 1) / 3
            If (intValidVectorCounter > (0.25 * intNumberOfRangers)) Then
                sngStdDeviation = Sqrt(sngStdDeviation / intValidVectorCounter)
                sngStdDeviation = ftnMaxValue(sngStdDeviation, 0.001)
                If boolKinectConnected Then
                    ftnRangerToMapCorrelation = intValidVectorCounter / sngStdDeviation * 2 / intNumberOfScanVectors
                Else
                    ftnRangerToMapCorrelation = intValidVectorCounter / sngStdDeviation / intNumberOfRangers
                End If
            End If
        End If
    End Function

    Function ftnGreySpace(ByVal tmcLocation As typeMapCoordinate) As Boolean
        ' return true if location is in Grey zone = cluttered environment
        On Error Resume Next
        Dim byteCellContent As Byte = byteNavigationMap(tmcLocation.X, tmcLocation.Y)
        ftnGreySpace = (byteCellContent > 32) And (byteCellContent < 128) And Not boolOccupiedCell(tmcLocation.X, tmcLocation.Y) And Not boolCompassIsSuspect(tmcLocation.X, tmcLocation.Y)
    End Function

    Function ftnPoseUncertaintyFrom2DScan(ByVal tbpCurrentPose As typeBotPose, _
                                          ByVal sngCorrelationFactor As Single, _
                                          ByVal sngStdDeviation As Single) As typeBotPose
        ' Reduce pose uncertainty if we have a good match        
        On Error Resume Next
        ftnPoseUncertaintyFrom2DScan = tbpCurrentPose
        If (sngCorrelationFactor >= sngExcellentCorrelation) Then
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X = ftnMinValue(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X, ftnInchesToMapUnits(sngStdDeviation * 8) / sngGridSize)
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y = ftnMinValue(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y, ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X)
            ftnPoseUncertaintyFrom2DScan.OrientationUncertainty = ftnMinValue(ftnPoseUncertaintyFrom2DScan.OrientationUncertainty, 10)
        ElseIf (sngCorrelationFactor >= sngGoodCorrelation) Then
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X = ftnMinValue(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X, ftnInchesToMapUnits(sngStdDeviation * 10) / sngGridSize)
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y = ftnMinValue(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y, ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X)
            ftnPoseUncertaintyFrom2DScan.OrientationUncertainty = ftnMinValue(ftnPoseUncertaintyFrom2DScan.OrientationUncertainty, 12)
        ElseIf (sngCorrelationFactor >= sngWeakCorrelation) Then
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X = ftnMinValue(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X, ftnInchesToMapUnits(sngStdDeviation * 14) / sngGridSize)
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y = ftnMinValue(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y, ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X)
            ftnPoseUncertaintyFrom2DScan.OrientationUncertainty = ftnMinValue(ftnPoseUncertaintyFrom2DScan.OrientationUncertainty, 16)
        End If
        ftnPoseUncertaintyFrom2DScan = ftnMinPoseUncertainty(ftnPoseUncertaintyFrom2DScan)
    End Function



    '--------------------------------------  Non Flat Wall  ----------------------------------------------------------------------

    Function ftnNonFlatWall() As Boolean
        ' long flat walls which fill the Kinect's FOV can cause localization errors
        On Error Resume Next
        ftnNonFlatWall = False

        Dim sngKinectRange(128) As Single
        Dim sngKinectSlope(128) As Single
        Dim intSamplesPerVector As Short = 128 / intNumberOfScanVectors
        Dim intTwoSamplesPerVector As Short = intSamplesPerVector + intSamplesPerVector
        For intKinectScanVector As Short = 0 To 127 Step intSamplesPerVector
            sngKinectRange(intKinectScanVector) = trvScanVectorsCopy(intKinectScanVector).sngRange
        Next intKinectScanVector
        If (sngKinectRange(intTwoSamplesPerVector) = 0) Then sngKinectRange(intTwoSamplesPerVector) = sngKinectRange(3 * intSamplesPerVector)
        If (sngKinectRange(intSamplesPerVector) = 0) Then sngKinectRange(intSamplesPerVector) = sngKinectRange(intTwoSamplesPerVector)
        If (sngKinectRange(0) = 0) Then sngKinectRange(0) = sngKinectRange(intSamplesPerVector)
        sngKinectSlope(intSamplesPerVector) = sngKinectRange(intSamplesPerVector) - sngKinectRange(0)
        Dim intI_MinusSamplesPerVector As Short
        For I As Short = intTwoSamplesPerVector To 127 Step intSamplesPerVector
            intI_MinusSamplesPerVector = I - intSamplesPerVector
            If (sngKinectRange(I) = 0) Then sngKinectRange(I) = sngKinectRange(intI_MinusSamplesPerVector) ' get rid of range = 0 entries
            sngKinectSlope(I) = sngKinectRange(I) - sngKinectRange(intI_MinusSamplesPerVector) ' calculate slopes
            If (sngKinectRange(I) <> 0) And (sngKinectRange(intI_MinusSamplesPerVector) <> 0) Then
                If (Abs(sngKinectRange(I) - sngKinectRange(intI_MinusSamplesPerVector)) > sngTwoInches) Then ' Exit if Spike (Discontinuity)
                    ftnNonFlatWall = True
                    Exit Function
                End If
            End If
        Next I
    End Function




End Module
