﻿
Imports System.Math

Module modLocalizeDeadReckoning


    Public Function ftnLocalizeUsingDeadReckoning(ByVal tbpEstimatedPose As typeBotPose, ByVal sngSensorDelay_seconds As Single, _
                                                  ByVal tbpTargetPose As typeBotPose) As typeBotPose
        ' sngSensorDelay_seconds is used to calculate the pose at the time sensor readings (Kinect, IR, camera, whatever) were measured
        On Error Resume Next
        Dim sngDistanceFromStart, sngAngleTraveled, sngElapsedTime As Single
        sngElapsedTime = ftnElapsedTime() - sngSensorDelay_seconds  ' in Seconds
        If (strTypeOfMove = "Turn") Then
            sngAngleTraveled = ftnElapsedAngle(sngElapsedTime)
            ftnLocalizeUsingDeadReckoning = ftnTurnDR(tbpStartingPose, sngAngleTraveled)
        ElseIf (strTypeOfMove = "Fwd") Then
            sngDistanceFromStart = ftnDistanceFromStart(sngElapsedTime)
            ftnLocalizeUsingDeadReckoning = ftnFwdDR(tbpStartingPose, sngDistanceFromStart)
        ElseIf (strTypeOfMove = "XY") Then
            sngDistanceFromStart = ftnDistanceFromStart(sngElapsedTime)
            ftnLocalizeUsingDeadReckoning = ftnXY_DR(tbpStartingPose, sngDistanceFromStart, tbpTargetPose, True)
        Else
            ftnLocalizeUsingDeadReckoning = tbpEstimatedPose
        End If
    End Function

    Public Function ftnFwdDR(ByVal tbpStartingPose As typeBotPose, ByVal sngDistanceFromStart As Single) As typeBotPose
        On Error Resume Next
        Dim sngAngleOfTurnRadians As Single = ftnDegreeToRadian(tbpStartingPose.Orientation)
        Dim intDeltaX As Integer = sngDistanceFromStart * System.Math.Cos(sngAngleOfTurnRadians) / sngGridSize
        Dim intDeltaY As Integer = sngDistanceFromStart * System.Math.Sin(-sngAngleOfTurnRadians) / sngGridSize
        ftnFwdDR.Orientation = ftnConvertTo360(tbpStartingPose.Orientation)
        ftnFwdDR = tbpStartingPose
        ftnFwdDR.Location.X += intDeltaX
        ftnFwdDR.Location.Y += intDeltaY
        Dim sngUncertaintyFactor As Single = sngBotMovementAccuracy * sngTenInches
        ftnFwdDR.LocationUncertainty.X = tbpStartingPose.LocationUncertainty.X + Abs(intDeltaX / sngUncertaintyFactor)
        ftnFwdDR.LocationUncertainty.Y = tbpStartingPose.LocationUncertainty.Y + Abs(intDeltaY / sngUncertaintyFactor)
        ftnFwdDR.OrientationUncertainty = tbpStartingPose.OrientationUncertainty + Abs(ftnMapUnitsToInches(sngDistanceFromStart) / (sngBotMovementAccuracy * 20))
    End Function



    Public Function ftnXY_DR(ByVal tbpStartingPose As typeBotPose, ByVal sngDistanceFromStart As Single, ByVal tbpTargetPose As typeBotPose, ByVal boolCalculateUncertainties As Boolean) As typeBotPose
        On Error Resume Next
        ' MANY Assumptions - XY Motion = drive forward while turning towards target, then drive straight toward target
        '                    Assumptions:  Radius of turn = diameter of Bot;  Speed = Constant for turning and forward motion;  
        ' see http://rossum.sourceforge.net/papers/CalculationsForRobotics/CirclePath.htm  (but his equations do not agree with mine)
        '
        ' RIGHT TURN
        ' X(t) = x(start) - radius * sin(tbpStartingPose.orientation) + radius *sin(Anglular Velocity*t+tbpStartingPose.orientation)
        ' Y(t) = y(start) + radius * cos(tbpStartingPose.orientation) - radius *cos(-Anglular Velocity*t+tbpStartingPose.orientation)
        '
        'LEFT TURN
        ' X(t) = x(center) - radius *sin(Anglular Velocity*t+tbpStartingPose.orientation-180)
        ' Y(t) = y(center) - radius *cos(Anglular Velocity*t+tbpStartingPose.orientation-180)
        ftnXY_DR = tbpStartingPose

        Dim sngAngleOfTurn As Single = ftnTurnAngle(tbpStartingPose, tbpTargetPose.Location)
        Dim boolTurnRight As Boolean = (sngAngleOfTurn < 0)
        sngAngleOfTurn = Abs(sngAngleOfTurn)
        Dim sngAngleOfTurnRadians As Single = ftnDegreeToRadian(sngAngleOfTurn)
        Dim sngXY_Radius As Single = sngBotDiameter '   ' BIG Assumption
        Dim sngXY_RadiusGrid As Single = sngXY_Radius / sngGridSize '
        Dim sngTurnDistance As Single = Abs(sngXY_Radius * sngAngleOfTurnRadians) ' NOTE: Simplified - no linear or angular velocities
        Dim sngTurnPercentage As Single = 1
        If (sngDistanceFromStart < sngTurnDistance) Then sngTurnPercentage = (sngDistanceFromStart / sngTurnDistance)
        Dim sngAngleOfTurnPercent As Single = sngAngleOfTurn * sngTurnPercentage
        Dim sngAngleOfTurnPercentRadians = (sngAngleOfTurnRadians * sngTurnPercentage)
        Dim StartOrientationRadians As Single = ftnConvertTo180(ftnDegreeToRadian(ftnXY_DR.Orientation))

        If boolTurnRight Then    ' Find the pose on the Turning portion of the move
            ftnXY_DR.Location.X = tbpStartingPose.Location.X + sngXY_RadiusGrid * (-Sin(StartOrientationRadians) + Sin(sngAngleOfTurnPercentRadians + StartOrientationRadians))
            ftnXY_DR.Location.Y = tbpStartingPose.Location.Y + sngXY_RadiusGrid * (Cos(StartOrientationRadians) - Cos(-sngAngleOfTurnPercentRadians + StartOrientationRadians))
            ftnXY_DR.Orientation = ftnConvertTo360(tbpStartingPose.Orientation - sngAngleOfTurnPercent)
        Else
            Dim tmcCenterOfTurn As typeMapCoordinate
            tmcCenterOfTurn = ftnLocationFromEstimatedPose(tbpStartingPose, 90, sngXY_Radius)
            ftnXY_DR.Location.X = tmcCenterOfTurn.X - sngXY_RadiusGrid * Sin(StartOrientationRadians + sngAngleOfTurnPercentRadians - 3.14159)
            ftnXY_DR.Location.Y = tmcCenterOfTurn.Y - sngXY_RadiusGrid * Cos(StartOrientationRadians + sngAngleOfTurnPercentRadians - 3.14159)   '  
            ftnXY_DR.Orientation = ftnConvertTo360(tbpStartingPose.Orientation + sngAngleOfTurnPercent)
        End If

        If (sngTurnPercentage >= 1) Then  ' Find the pose on the Straight portion of the move
            ftnXY_DR.Orientation = ftnP1ToP2Angle(ftnXY_DR.Location, tbpTargetPose.Location)
            Dim sngDistanceFromCurveEndToTarget As Single = ftnDistanceBetweenLocations(ftnXY_DR.Location, tbpTargetPose.Location)
            ftnXY_DR.Location = ftnLocationFromEstimatedPose(ftnXY_DR, 0, ftnMinValue((sngDistanceFromStart - sngTurnDistance), sngDistanceFromCurveEndToTarget))
        End If

        If boolCalculateUncertainties Then
            Dim intDeltaX As Integer = ftnXY_DR.Location.X - tbpStartingPose.Location.X
            Dim intDeltaY As Integer = ftnXY_DR.Location.Y - tbpStartingPose.Location.Y
            Dim sngUncertaintyFactor As Single = sngBotMovementAccuracy * sngTenInches
            ftnXY_DR.LocationUncertainty.X = tbpStartingPose.LocationUncertainty.X + Abs(intDeltaX / sngUncertaintyFactor)
            ftnXY_DR.LocationUncertainty.Y = tbpStartingPose.LocationUncertainty.Y + Abs(intDeltaY / sngUncertaintyFactor)
            ftnXY_DR.OrientationUncertainty = ftnMinValue(16, tbpStartingPose.OrientationUncertainty + Abs(ftnAngleDifference(tbpStartingPose.Orientation, ftnXY_DR.Orientation)) / (sngBotMovementAccuracy * 10))
            If ftnMapLocationsEqual(ftnXY_DR.Location, tbpTargetPose.Location) Then ftnXY_DR.Orientation = tbpTargetPose.Orientation
        End If
    End Function


    Function ftnTurnDR(ByVal tbpStartingPose As typeBotPose, ByVal sngTurnAngle As Single) As typeBotPose
        On Error Resume Next
        ftnTurnDR = tbpStartingPose
        ftnTurnDR.Orientation = ftnConvertTo360(tbpStartingPose.Orientation + sngTurnAngle)
        ftnTurnDR.OrientationUncertainty += Abs(ftnConvertTo180(sngTurnAngle) / (sngBotMovementAccuracy * 20))
        Dim sngLocationUncertaintyIncrease As Single = sngOneInch / sngGridSize
        ftnTurnDR.LocationUncertainty.X += sngLocationUncertaintyIncrease
        ftnTurnDR.LocationUncertainty.Y += sngLocationUncertaintyIncrease
    End Function

    Public Function ftnElapsedTime() As Single ' returns elapsed time since start of move in seconds
        On Error Resume Next
        Dim tsElapsedTimeSpan As TimeSpan = Now().Subtract(timeMoveCommandSent)
        ftnElapsedTime = (Convert.ToSingle(tsElapsedTimeSpan.TotalMilliseconds.ToString) - sngStartOfMoveDelay) * 0.001
        If (ftnElapsedTime < 0) Then ftnElapsedTime = 0
    End Function


    Function ftnDistanceFromStart(ByVal sngElapsedTime As Single) As Single
        ' d = vt + (1/2)at^2 
        ' v = a t  
        On Error Resume Next
        Dim sngDistanceCommanded As Single = ftnDistanceCommanded(tbpStartingPose, tbpTargetPoseGlobal)
        If (strTypeOfMove = "XY") Then
            Dim sngAngleOfTurn As Single = ftnTurnAngle(tbpStartingPose, tbpTargetPoseGlobal.Location)
            sngDistanceCommanded = sngDistanceCommanded * 2  '/ Cos(ftnDegreeToRadian(sngAngleOfTurn)) ' 
        End If
        If (sngElapsedTime > sngBotLinearAccelerationTime) Then
            If (strTypeOfMove = "XY") Then
                ftnDistanceFromStart = sngBotLinearAccelerationDistanceXY + ftnCmToMapUnits((sngElapsedTime - sngBotLinearAccelerationTimeXY) * sngBotLinearVelocityXY)
            Else
                ftnDistanceFromStart = sngBotLinearAccelerationDistance + ftnCmToMapUnits((sngElapsedTime - sngBotLinearAccelerationTime) * sngBotLinearVelocity)
            End If
        Else ' TODO Add deacceleration distance
            If (strTypeOfMove = "XY") Then
                ftnDistanceFromStart = ftnCmToMapUnits(sngBotLinearAcceleration * sngElapsedTime * sngElapsedTime * 0.5)
            Else
                ftnDistanceFromStart = ftnCmToMapUnits(sngBotLinearAcceleration * sngElapsedTime * sngElapsedTime * 0.5)
            End If
        End If
        If (ftnDistanceFromStart < 0) Then ftnDistanceFromStart = 0
        If (ftnDistanceFromStart > Abs(sngDistanceCommanded)) Then ftnDistanceFromStart = Abs(sngDistanceCommanded)
        If (sngDistanceCommanded < 0) Then ftnDistanceFromStart = -ftnDistanceFromStart
    End Function

    Function ftnElapsedAngle(ByVal sngElapsedTime As Single) As Single
        On Error Resume Next
        Dim sngTurnAngleCommanded As Single = ftnAngleDifference(tbpTargetPoseGlobal.Orientation, tbpStartingPose.Orientation)
        If (sngElapsedTime > sngBotAngularAccelerationTime) Then
            ftnElapsedAngle = sngBotAngularAccelerationDistance + (sngElapsedTime - sngBotAngularAccelerationTime) * sngBotAngularVelocity
        Else
            ftnElapsedAngle = sngBotAngularAcceleration * sngElapsedTime * sngElapsedTime * 0.5
        End If
        If (ftnElapsedAngle < 0) Then ftnElapsedAngle = 0
        If (ftnElapsedAngle > Abs(sngTurnAngleCommanded)) Then ftnElapsedAngle = Abs(sngTurnAngleCommanded)
        If (sngTurnAngleCommanded < 0) Then ftnElapsedAngle = -ftnElapsedAngle
    End Function



End Module






'Dim sngTurnDistance As Single = Abs(sngBotRadius * sngAngleOfTurnRadians) ' NOTE: Simplified - not linear or angular velocities

'Dim tmcCenterOfTurn As typeMapCoordinate
'If (sngAngleOfTurn < 0) Then  ' Find the center point of the Turning portion of the move 
'    tmcCenterOfTurn = ftnLocationFromEstimatedPose(tbpStartingPose, -90, sngXY_Radius)
'Else
'    tmcCenterOfTurn = ftnLocationFromEstimatedPose(tbpStartingPose, 90, sngXY_Radius)
'End If
'If boolBotIsMoving Then subDrawSmallPurpleCircle(tmcCenterOfTurn)

'Dim sngTurnPercentage As Single = 1
'If (sngDistanceFromStart < sngTurnDistance) Then sngTurnPercentage = (sngDistanceFromStart / sngTurnDistance) ^ 2 ' ???  Percentage is squared so as to mimic the sinusoidal angle change
'sngAngleOfTurn = sngAngleOfTurn * sngTurnPercentage
'sngAngleOfTurnRadians = sngAngleOfTurnRadians * sngTurnPercentage

'ftnXY_DR.Orientation = ftnConvertTo360(tbpStartingPose.Orientation + sngAngleOfTurn) ' Find the pose on the Turning portion of the move
'Dim boolOrientationPointsRight As Boolean = ftnOrientationPointsRight(ftnXY_DR.Orientation)
'If boolOrientationPointsRight Then
'    ftnXY_DR.Location.Y = tbpStartingPose.Location.Y - sngXY_RadiusGrid * Sin(sngAngleOfTurnRadians)
'Else
'    ftnXY_DR.Location.Y = tbpStartingPose.Location.Y + sngXY_RadiusGrid * Sin(sngAngleOfTurnRadians)
'End If

'Dim integerXOffset As Short = Sqrt(sngXY_RadiusGrid ^ 2 - (tmcCenterOfTurn.Y - ftnXY_DR.Location.Y) ^ 2) + 0.5
'ftnXY_DR.Location.X = tmcCenterOfTurn.X + integerXOffset
'If (Abs(ftnAngleDifference(tbpStartingPose.Orientation, ftnP1ToP2Angle(tbpStartingPose.Location, ftnXY_DR.Location))) > sngAngleOfTurn + 10) Then ' SO UGLY
'    ftnXY_DR.Location.X = tmcCenterOfTurn.X - integerXOffset
'End If
'If boolOrientationPointsRight And (sngAngleOfTurnRadians > 0) Then
'    ftnXY_DR.Location.X = tmcCenterOfTurn.X + integerXOffset
'ElseIf boolOrientationPointsRight Then
'    ftnXY_DR.Location.X = tbpStartingPose.Location.X + Sqrt(sngXY_RadiusGrid ^ 2 - (tmcCenterOfTurn.X = ftnXY_DR.Location.X) ^ 2)
'End If

'If (sngTurnPercentage >= 1) Then  ' Find the pose on the Straight portion of the move
'    If boolBotIsMoving Then subDrawSmallRedCircle(ftnXY_DR.Location)
'    ftnXY_DR.Orientation = ftnP1ToP2Angle(ftnXY_DR.Location, tbpTargetPose.Location)
'    Dim sngDistanceFromCurveEndToTarget As Single = ftnDistanceBetweenLocations(ftnXY_DR.Location, tbpTargetPose.Location)
'    ftnXY_DR.Location = ftnLocationFromEstimatedPose(ftnXY_DR, 0, ftnMinValue((sngDistanceFromStart - sngTurnDistance), sngDistanceFromCurveEndToTarget))
'End If
